"""Raw features from FRED sources."""

import logging

import pandas as pd
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound
from tqdm import tqdm

from ... import backend, utils
from .. import api, sql

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class Series:
    """Get a single economic series as-is from raw FRED data.

    The module variable :data:`finagg.fred.feat.series` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    """

    @classmethod
    def from_raw(
        cls,
        series_id: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get a single economic data series as-is from raw FRED data.

        This is the preferred method for accessing raw FRED data without
        using the FRED API.

        Args:
            series_id: Economic data series ID.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            A dataframe containing the economic data series values
            across the specified period.

        Raises:
            `NoResultFound`: If there are no rows for ``series_id`` in the
                raw SQL table.

        Examples:
            >>> finagg.fred.feat.series.from_raw("CPIAUCNS").head(5)  # doctest: +SKIP
                        value
            date
            1949-03-01  169.5
            1949-04-01  169.7
            1949-05-01  169.2
            1949-06-01  169.6
            1949-07-01  168.5

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.series.name):
            sql.series.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sa.select(sql.series.c.date, sql.series.c.value).where(
                        sql.series.c.series_id == series_id,
                        sql.series.c.date >= start,
                        sql.series.c.date <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No series rows found for {series_id}.")
        return df.set_index("date").sort_index()

    @classmethod
    def get_id_set(cls, lb: int = 1, *, engine: None | Engine = None) -> set[str]:
        """Get all unique economic series IDs in the raw SQL tables that have at least
        ``lb`` rows.

        Args:
            lb: Lower bound number of rows that a series must have for its ID
                to be included in the set returned by this method.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Examples:
            >>> "FEDFUNDS" in finagg.fred.feat.series.get_id_set()  # doctest: +SKIP
            True

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.series.name):
            sql.series.create(engine)
        with engine.begin() as conn:
            series_ids = (
                conn.execute(
                    sa.select(sql.series.c.series_id)
                    .group_by(sql.series.c.series_id)
                    .having(sa.func.count(sql.series.c.date) >= lb)
                )
                .scalars()
                .all()
            )
        return set(series_ids)

    @classmethod
    def install(
        cls,
        series_ids: None | set[str] = None,
        *,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with by pulling data from the FRED API and
        then writing the data to the raw series SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            series_ids: Set of series to install features for. Defaults to all
                the series from :attr:`finagg.fred.feat.Series.get_id_set`.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        series_ids = series_ids or set(api.popular_series)
        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.series.name):
            sql.series.drop(engine, checkfirst=True)
            sql.series.create(engine)

        total_rows = 0
        for series_id in tqdm(
            series_ids,
            desc="Installing raw FRED economic series data",
            position=0,
            leave=True,
        ):
            try:
                df = api.series.observations.get_original_observations(
                    series_id,
                )
                rowcount = len(df.index)
                if rowcount:
                    cls.to_raw(df, engine=engine)
                    total_rows += rowcount
                    logger.debug(f"{rowcount} rows inserted for {series_id}")
                else:
                    logger.debug(f"Skipping {series_id} due to missing data")
            except Exception as e:
                logger.debug(f"Skipping {series_id}", exc_info=e)
        return total_rows

    @classmethod
    def to_raw(cls, df: pd.DataFrame, /, *, engine: None | Engine = None) -> int:
        """Write the given dataframe to the raw feature table.

        Args:
            df: Dataframe to store as rows in a local SQL table
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.series.name):
            sql.series.create(engine)
        with engine.begin() as conn:
            conn.execute(sql.series.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df)



"""BEA SQLAlchemy interfaces."""


from sqlalchemy import Column, Float, Integer, MetaData, String, Table

metadata = MetaData()
"""The metadata associated with all SQL tables defined in this module.

:meta hide-value:
"""

fixed_assets = Table(
    "fixed_assets",
    metadata,
    Column("table_id", String, primary_key=True),
    Column("series_code", String),
    Column("line", Integer, primary_key=True),
    Column("line_description", String),
    Column("year", Integer, primary_key=True),
    Column("metric", String),
    Column("units", String),
    Column("e", Integer),
    Column("value", Float),
)
"""SQL table for storing raw data as returned by
:data:`finagg.bea.api.fixed_assets` (an alias for
:class:`finagg.bea.api.FixedAssets`).

:meta hide-value:
"""

gdp_by_industry = Table(
    "gdp_by_industry",
    metadata,
    Column("table_id", Integer, primary_key=True),
    Column("freq", String),
    Column("year", Integer, primary_key=True),
    Column("quarter", Integer, primary_key=True),
    Column("industry", String, primary_key=True),
    Column("industry_description", String),
    Column("value", Float),
)
"""SQL table for storing raw data as returned by
:data:`finagg.bea.api.gdp_by_industry` (an alias for
:class:`finagg.bea.api.GDPByIndustry`).

:meta hide-value:
"""

input_output = Table(
    "input_output",
    metadata,
    Column("table_id", Integer, primary_key=True),
    Column("year", Integer, primary_key=True),
    Column("row_code", String, primary_key=True),
    Column("row_description", String),
    Column("row_type", String),
    Column("col_code", String, primary_key=True),
    Column("col_description", String),
    Column("col_type", String),
    Column("value", Float),
)
"""SQL table for storing raw data as returned by
:data:`finagg.bea.api.input_output` (an alias for
:class:`finagg.bea.api.InputOutput`).

:meta hide-value:
"""

nipa = Table(
    "nipa",
    metadata,
    Column("table_id", String, primary_key=True),
    Column("series_code", String),
    Column("line", Integer, primary_key=True),
    Column("line_description", String),
    Column("year", Integer, primary_key=True),
    Column("quarter", Integer, primary_key=True),
    Column("metric", String),
    Column("units", String),
    Column("e", Integer),
    Column("value", Float),
)
"""SQL table for storing raw data as returned by
:data:`finagg.bea.api.nipa` (an alias for
:class:`finagg.bea.api.NIPA`).

:meta hide-value:
"""



from typing import Generator

import pytest
from sqlalchemy.engine import Engine

import finagg


@pytest.fixture
def engine() -> Generator[Engine, None, None]:
    yield from finagg.testing.sqlite_engine(
        finagg.backend.database_path, metadata=finagg.sec.sql.metadata
    )


def test_get_cik(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    assert finagg.sec.sql.get_cik("AAPL", engine=engine) == "0000320193"


def test_get_metadata(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    assert (
        finagg.sec.sql.get_metadata(ticker="AAPL", engine=engine)["cik"] == "0000320193"
    )


def test_get_ticker(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    assert finagg.sec.sql.get_ticker("0000320193", engine=engine) == "AAPL"


def test_get_ticker_set(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"AAPL"}, engine=engine)
    finagg.sec.feat.tags.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.sec.feat.tags.get_ticker_set(engine=engine)


def test_get_tickers_in_industry(engine: Engine) -> None:
    finagg.sec.feat.submissions.install({"HD", "LOW"}, engine=engine)
    assert "LOW" in finagg.sec.sql.get_tickers_in_industry(ticker="HD", engine=engine)



"""SEC CLI and tools."""

import logging
import multiprocessing as mp
import os
from typing import Literal

import click

from .. import indices, utils
from . import api as _api
from . import feat as _feat

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


@click.group(help="Securities and Exchange Commission (SEC) tools.")
def entry_point() -> None:
    ...


@entry_point.command(
    help=(
        "Set the SEC API key, drop and recreate tables, "
        "and install the recommended tables into the SQL database."
    ),
)
@click.option(
    "--raw",
    type=click.Choice(["submissions", "tags"]),
    multiple=True,
    help=(
        "Raw tables to install. `submissions` indicates company metadata "
        "(e.g., company name, industry code, etc.) while `tags` indicates "
        "SEC EDGAR tags (e.g., earnings-per-share, current assets, etc.)."
        "Both `submissions` and `tags` must be specified to enable installing "
        "refined data using the `refined` flag."
    ),
)
@click.option(
    "--refined",
    type=click.Choice(
        ["annual", "annual.normalized", "quarterly", "quarterly.normalized"]
    ),
    multiple=True,
    help=(
        "Refined tables to install. This requires raw SEC data to be "
        "installed beforehand using the `--raw` flag or for the "
        "`--raw` flag to be set when this option is provided."
    ),
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    default=False,
    help="Whether to install all defined tables (including all refined tables).",
)
@click.option(
    "--ticker",
    "-t",
    multiple=True,
    help=(
        "Ticker whose data is attempted to be downloaded and inserted into "
        "the SQL tables. Multiple tickers can be specified by providing "
        "multiple `ticker` options, by separating tickers with a comma (e.g., "
        "`AAPL,MSFT,NVDA`), or by providing tickers in a CSV file by "
        "specifying a file path (e.g., `bank_tickers.txt`). The CSV file "
        "can be formatted such that there's one ticker per line or multiple "
        "tickers per line (delimited by a comma). The tickers specified "
        "by this option are combined with the tickers specified by the "
        "`ticker-set` option."
    ),
)
@click.option(
    "--ticker-set",
    "-ts",
    "ticker_set",
    type=click.Choice(["indices", "sec"]),
    default=None,
    help=(
        "Set of tickers whose data is attempted to be downloaded and "
        "inserted into the SQL tables. 'indices' indicates the set "
        "of tickers from the three most popular indices (DJIA, "
        "Nasdaq 100, and S&P 500). 'sec' indicates all the tickers that "
        "have data available through the SEC API (which is approximately "
        "all publicly-traded US companies). 'indices' will effectively "
        "only attempt to download and install data for relatively "
        "popular and large market cap companies, while 'sec' will "
        "attempt to download and install data for nearly all "
        "publicly-traded US companies. Choosing 'indices' will be fast, "
        "while choosing 'sec' will be slow but will include more diverse data."
    ),
)
@click.option(
    "--from-zip",
    "-z",
    is_flag=True,
    default=False,
    help=(
        "Whether to install raw data from bulk data zip files that're compiled by the"
        " SEC nightly. Installing all SEC data with this option can take upwards of 1.5"
        " hours to complete."
    ),
)
@click.option(
    "--processes",
    "-n",
    type=int,
    default=mp.cpu_count() - 1,
    help=(
        "Number of backgruond processes to run in parallel when installing data. Note,"
        " not all tables support installations with multiprocessing."
    ),
)
@click.option(
    "--recreate-tables",
    "-r",
    is_flag=True,
    default=False,
    help=(
        "Whether to reset the tables associated with the install options by "
        "dropping and recreating them."
    ),
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Sets the log level to DEBUG to show installation errors for each ticker.",
)
def install(
    raw: list[Literal["submissions", "tags"]] = [],
    refined: list[Literal["quarterly", "quarterly.normalized"]] = [],
    all_: bool = False,
    ticker: list[str] = [],
    ticker_set: None | Literal["indices", "sec"] = None,
    from_zip: bool = False,
    processes: int = mp.cpu_count() - 1,
    recreate_tables: bool = False,
    verbose: bool = False,
) -> int:
    if verbose:
        logging.getLogger(__package__).setLevel(logging.DEBUG)

    if "SEC_API_USER_AGENT" not in os.environ:
        user_agent = input(
            "Enter your SEC API user agent below (this should be of format "
            "'FIRST_NAME LAST_NAME E_MAIL').\n\n"
            "SEC API user agent (leave blank and hit ENTER to skip): "
        ).strip()
        if not user_agent:
            logger.warning(
                "An empty SEC API user agent was given. Skipping finagg.sec "
                "installation."
            )
            return 0
        p = utils.setenv("SEC_API_USER_AGENT", user_agent)
        logger.info(f"SEC API user agent written to {p}")
    else:
        logger.info("SEC API user agent found in the environment")

    total_rows = 0
    all_raw = set()
    if all_:
        all_raw = {"submissions", "tags"}
    elif raw:
        all_raw = set(raw)

    all_tickers = utils.expand_csv(ticker)
    if all_raw:
        match ticker_set:
            case "indices":
                all_tickers |= indices.api.get_ticker_set()
            case "sec":
                all_tickers |= _api.get_ticker_set()

        if not all_tickers:
            logger.info(
                f"Skipping {__package__} installation because no tickers were "
                "provided (by the `ticker` option or by the `ticker-set` option)"
            )
            return total_rows

        if "submissions" in all_raw:
            if from_zip:
                total_rows += _feat.submissions.install_from_zip(
                    all_tickers, recreate_tables=recreate_tables
                )
            else:
                total_rows += _feat.submissions.install(
                    all_tickers, recreate_tables=recreate_tables
                )

        if "tags" in all_raw:
            if from_zip:
                total_rows += _feat.tags.install_from_zip(
                    all_tickers, processes=processes, recreate_tables=recreate_tables
                )
            else:
                total_rows += _feat.tags.install(
                    all_tickers, recreate_tables=recreate_tables
                )

    all_refined = set()
    if all_:
        all_refined = {
            "annual",
            "annual.normalized",
            "quarterly",
            "quarterly.normalized",
        }
    elif refined:
        all_refined = set(refined)

    if "annual" in all_refined:
        total_rows += _feat.annual.install(
            tickers=all_tickers, processes=processes, recreate_tables=recreate_tables
        )

    if "annual.normalized" in all_refined:
        total_rows += _feat.annual.normalized.install(
            tickers=all_tickers, processes=processes, recreate_tables=recreate_tables
        )

    if "quarterly" in all_refined:
        total_rows += _feat.quarterly.install(
            tickers=all_tickers, processes=processes, recreate_tables=recreate_tables
        )

    if "quarterly.normalized" in all_refined:
        total_rows += _feat.quarterly.normalized.install(
            tickers=all_tickers, processes=processes, recreate_tables=recreate_tables
        )

    if all_ or all_refined or all_raw:
        if total_rows:
            logger.info(f"{total_rows} total rows inserted for {__package__}")
        else:
            logger.warning(
                f"No rows were inserted for {__package__}. This could be an error if"
                " installations were skipped. Set the verbose flag with the"
                " `--verbose/-v` option to enable debug logging."
            )
    else:
        logger.info(
            f"Skipping {__package__} installation because no installation "
            "options were provided"
        )
    return total_rows



from datetime import datetime, timedelta
from typing import Generator

import pandas as pd
import pytest
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError

import finagg


@pytest.fixture
def engine() -> Generator[Engine, None, None]:
    yield from finagg.testing.sqlite_engine(
        finagg.backend.database_path, table=finagg.yfinance.sql.daily
    )


def test_daily_all_equal(engine: Engine) -> None:
    finagg.yfinance.feat.prices.install({"AAPL"}, engine=engine)
    finagg.yfinance.feat.daily.install({"AAPL"}, engine=engine)
    df1 = finagg.yfinance.feat.daily.from_api("AAPL").head(5)
    df2 = finagg.yfinance.feat.daily.from_raw("AAPL", engine=engine).head(5)
    df3 = finagg.yfinance.feat.daily.from_refined("AAPL", engine=engine).head(5)
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
    pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)


def test_daily_get_candidate_ticker_set(engine: Engine) -> None:
    finagg.yfinance.feat.prices.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.yfinance.feat.prices.get_ticker_set(engine=engine)
    assert "AAPL" in finagg.yfinance.feat.daily.get_candidate_ticker_set(engine=engine)


def test_daily_get_candidate_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.yfinance.feat.daily.get_candidate_ticker_set(engine=engine)) == 0


def test_daily_get_ticker_set(engine: Engine) -> None:
    finagg.yfinance.feat.prices.install({"AAPL"}, engine=engine)
    finagg.yfinance.feat.daily.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.yfinance.feat.daily.get_candidate_ticker_set(engine=engine)
    assert "AAPL" in finagg.yfinance.feat.daily.get_ticker_set(engine=engine)


def test_daily_get_ticker_set_empty(engine: Engine) -> None:
    assert len(finagg.yfinance.feat.daily.get_ticker_set(engine=engine)) == 0


def test_daily_to_from_refined(engine: Engine) -> None:
    df1 = finagg.yfinance.feat.daily.from_api("AAPL")
    finagg.yfinance.feat.daily.to_refined("AAPL", df1, engine=engine)
    with pytest.raises(IntegrityError):
        finagg.yfinance.feat.daily.to_refined("AAPL", df1, engine=engine)

    df2 = finagg.yfinance.feat.daily.from_refined("AAPL", engine=engine)
    pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)


def test_daily_update(engine: Engine) -> None:
    install_end = datetime.fromisoformat(finagg.utils.today) - timedelta(days=14)
    finagg.yfinance.feat.prices.install({"AAPL"}, engine=engine)
    df = finagg.yfinance.feat.daily.from_api(
        "AAPL", end=install_end.strftime("%Y-%m-%d")
    )
    finagg.yfinance.feat.daily.to_refined("AAPL", df, engine=engine)
    finagg.yfinance.feat.daily.update({"AAPL"}, engine=engine)
    update_end = install_end + timedelta(days=7)
    df1 = finagg.yfinance.feat.daily.from_api(
        "AAPL", end=update_end.strftime("%Y-%m-%d")
    )
    df2 = finagg.yfinance.feat.daily.from_refined(
        "AAPL", end=update_end.strftime("%Y-%m-%d"), engine=engine
    )
    pd.testing.assert_frame_equal(df1, df2, atol=1e-4)


def test_prices_get_ticker_set(engine: Engine) -> None:
    finagg.yfinance.feat.prices.install({"AAPL"}, engine=engine)
    assert "AAPL" in finagg.yfinance.feat.prices.get_ticker_set(engine=engine)


def test_prices_update(engine: Engine) -> None:
    install_end = datetime.fromisoformat(finagg.utils.today) - timedelta(days=14)
    df = finagg.yfinance.api.get("AAPL", end=install_end.strftime("%Y-%m-%d"))
    finagg.yfinance.feat.prices.to_raw(df, engine=engine)
    finagg.yfinance.feat.prices.update({"AAPL"}, engine=engine)
    update_end = install_end + timedelta(days=7)
    df1 = finagg.yfinance.api.get("AAPL", end=update_end.strftime("%Y-%m-%d"))
    df2 = finagg.yfinance.feat.prices.from_raw(
        "AAPL", end=update_end.strftime("%Y-%m-%d"), engine=engine
    ).reset_index()
    df2["ticker"] = "AAPL"
    pd.testing.assert_frame_equal(df1, df2, atol=1e-4)



""":mod:`finagg` configuration and global SQLAlchemy setup. Backend file paths
and SQLAlchemy engine database URLs are configured in this module at runtime
according to environment variables.

Environment variables should ideally be configured using an ``.env`` file
in the desired working directory. Running ``finagg install`` will
automaticaly setup the ``.env`` file for you according to your input values.
Environment variables assigned in the ``.env`` file are loaded on the
:mod:`finagg` module's first instantiation.

"""

import os
import pathlib

from sqlalchemy import create_engine

root_path = pathlib.Path(os.environ.get("FINAGG_ROOT_PATH", pathlib.Path.cwd()))
"""Parent directory of the ``findata`` directory where the backend database
and API cache file will be stored (unless otherwise configured according
to the relevant environment variables). This can be set with the
``FINAGG_ROOT_PATH`` environment variable. This defaults to and is typically
set to the current working directory. It's recommended you permanently set
this value using the ``finagg install`` CLI.

:meta hide-value:
"""

disable_http_cache = os.environ.get("FINAGG_DISABLE_HTTP_CACHE", "false").lower() in {
    "1",
    "true",
}
"""Whether the disable the HTTP requests cache. Instead of a cachable session,
a default, uncached user session will be used for all requests.

:meta hide-value:
"""

http_cache_path = pathlib.Path(
    os.environ.get("FINAGG_HTTP_CACHE_PATH", root_path / "findata" / "http_cache")
)
"""Path to the API cache file. This can be set with the
``FINAGG_HTTP_CACHE_PATH`` environment variable and should NOT include a file
extension. All API implementations share the same cache backend.

:meta hide-value:
"""

database_path = root_path / "findata" / "finagg.sqlite"
"""Default path to the database file. The ``FINAGG_DATABASE_URL`` environment
variable will take precedence over this value.

:meta hide-value:
"""

database_url = os.environ.get("FINAGG_DATABASE_URL", f"sqlite:///{database_path}")
"""SQLAlchemy URL to the database. This can be set with the
``FINAGG_DATABASE_URL`` environment variable and should include a file extension.
This defaults to ``f"sqlite:///{finagg.backend.database_path}"``.

:meta hide-value:
"""

engine = create_engine(database_url)
"""The default SQLAlchemy engine for the backend database. All feature and SQL
submodules use this engine and the database URL as configured by
:data:`database_url` for reading and writing to and from the database by default.

:meta hide-value:
"""



"""Yahoo! finance SQLAlchemy interfaces."""

import sqlalchemy as sa

metadata = sa.MetaData()
"""The metadata associated with all SQL tables defined in this module.

:meta hide-value:
"""

prices = sa.Table(
    "yfinance.raw.prices",
    metadata,
    sa.Column("ticker", sa.String, primary_key=True, doc="Unique company ticker."),
    sa.Column("date", sa.String, primary_key=True, doc="Stock price date."),
    sa.Column("open", sa.Float, doc="Stock price at market open."),
    sa.Column("high", sa.Float, doc="Stock price max during trading hours."),
    sa.Column("low", sa.Float, doc="Stock price min during trading hours."),
    sa.Column("close", sa.Float, doc="Stock price at market close."),
    sa.Column("volume", sa.Integer, doc="Units traded during trading hours."),
)
"""SQL table for storing raw data as managed by
:data:`finagg.yfinance.feat.prices` (an alias for
:class:`finagg.yfinance.feat.Prices`).

:meta hide-value:
"""

daily = sa.Table(
    "yfinance.refined.daily",
    metadata,
    sa.Column(
        "ticker",
        sa.String,
        sa.ForeignKey(prices.c.ticker, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique company ticker.",
    ),
    sa.Column("date", sa.String, primary_key=True, doc="Date associated with feature."),
    sa.Column(
        "LOG_CHANGE(open)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in stock price at market open between the "
            "row's date and the previous date."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(high)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in stock price max during trading hours "
            "between the row's date and the previous date."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(low)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in stock price min during trading hours "
            "between the row's date and the previous date."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(close)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in stock price at market close between "
            "the row's date and the previous date."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(volume)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in stock units traded between the row's "
            "date and the previous date."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(high, open)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change between the stock price at market open "
            "and the stock price max during trading hours."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(low, open)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change between the stock price at market open "
            "and the stock price min during trading hours."
        ),
    ),
)
"""SQL table for storing refined data as managed by
:data:`finagg.yfinance.feat.daily` (an alias for
:class:`finagg.yfinance.feat.Daily`).

:meta hide-value:
"""



"""Annual features from SEC sources."""

import logging
import multiprocessing as mp
from typing import Literal

import numpy as np
import pandas as pd
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound

from .... import backend, utils
from ... import api, sql
from .. import _raw

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class IndustryAnnual:
    """Methods for gathering industry-averaged annual data from SEC
    features.

    The class variable :attr:`finagg.sec.feat.Annual.industry` is an
    instance of this feature set implementation and is the most popular
    interface for calling feature methods.

    Examples:
        You can aggregate this feature set using a ticker or an industry code
        directly.

        >>> df1 = finagg.sec.feat.annual.industry.from_refined(ticker="MSFT").head(5)
        >>> df2 = finagg.sec.feat.annual.industry.from_refined(code=73).head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)

    """

    @classmethod
    def from_refined(
        cls,
        /,
        *,
        ticker: None | str = None,
        code: None | str = None,
        level: Literal[2, 3, 4] = 2,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get annual features from the feature store,
        aggregated for an entire industry.

        The industry can be chosen according to a company or
        by an industry code directly. If a company is provided,
        then the first ``level`` digits of the company's SIC code
        is used for the industry code.

        Args:
            ticker: Company ticker. Lookup the industry associated
                with this company. Mutually exclusive with ``code``.
            code: Industry SIC code to use for industry lookup.
                Mutually exclusive with ``ticker``.
            level: Industry level to aggregate features at.
                The industry used according to ``ticker`` or ``code``
                is subsampled according to this value. Options include:

                    - 2 = major group (e.g., furniture and fixtures)
                    - 3 = industry group (e.g., office furnitures)
                    - 4 = industry (e.g., wood office furniture)

            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Annual data dataframe with each tag as a
            separate column. Sorted by filing date.

        Raises:
            `ValueError`: If neither a ``ticker`` nor ``code`` are provided.
            `NoResultFound`: If there are no rows for ``ticker`` or ``code``
                in the refined SQL table.

        Examples:
            >>> df = finagg.sec.feat.annual.industry.from_refined(ticker="AAPL").head(5)
            >>> df["mean"]  # doctest: +SKIP
            name             AssetCoverageRatio  DebtEquityRatio  EarningsPerShareBasic ...
            fy   filed                                                                  ...
            2010 2011-02-24            1.577924         1.112079               3.630000 ...
            2011 2012-02-24            1.998633         0.886148               1.851429 ...
            2012 2013-02-26            1.997748         1.413677               3.750833 ...
            2013 2014-03-03            2.232804         1.176639               4.475000 ...
            2014 2015-03-17            2.019007         1.080055               2.742143 ...
            >>> df["std"]  # doctest: +SKIP
            name             AssetCoverageRatio  DebtEquityRatio  EarningsPerShareBasic ...
            fy   filed                                                                  ...
            2010 2011-02-24            0.320778         0.149843               3.436292 ...
            2011 2012-02-24            1.371792         0.331815               3.837792 ...
            2012 2013-02-26            1.480323         1.085814               3.975254 ...
            2013 2014-03-03            1.787131         0.575039               7.236695 ...
            2014 2015-03-17            1.303284         0.458704               3.015323 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.annual.name):
            sql.annual.create(engine)
        with engine.begin() as conn:
            if ticker:
                (sic,) = conn.execute(
                    sa.select(sql.submissions.c.sic).where(
                        sql.submissions.c.ticker == ticker
                    )
                ).one()
                code = str(sic)[:level]
            elif code:
                code = str(code)[:level]
            else:
                raise ValueError("Must provide a `ticker` or `code`.")

            df = pd.DataFrame(
                conn.execute(
                    sql.annual.select()
                    .join(
                        sql.submissions,
                        (sql.submissions.c.cik == sql.annual.c.cik)
                        & (sql.submissions.c.sic.startswith(code)),
                    )
                    .where(sql.annual.c.filed >= start, sql.annual.c.filed <= end)
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No industry annual rows found for industry {code}.")
        df = df.drop(columns=["cik"])
        df = df.melt(["fy", "filed"], var_name="name", value_name="value").set_index(
            ["fy"]
        )
        df["filed"] = df.groupby(["fy"])["filed"].max()
        return (
            df.reset_index()  # type: ignore[return-value]
            .set_index(["fy", "filed"])
            .groupby(["fy", "filed", "name"])
            .agg(["mean", "std"])
            .reset_index()
            .pivot(index=["fy", "filed"], columns="name")["value"]
            .sort_index()
            .dropna()
        )


class NormalizedAnnual:
    """Annual features from SEC EDGAR data normalized according to industry
    averages and standard deviations.

    The class variable :attr:`finagg.sec.feat.Annual.normalized` is an
    instance of this feature set implementation and is the most popular
    interface for calling feature methods.

    Examples:
        It doesn't matter which data source you use to gather features.
        They both return equivalent dataframes.

        >>> df1 = finagg.sec.feat.annual.normalized.from_other_refined("AAPL").head(5)
        >>> df2 = finagg.sec.feat.annual.normalized.from_refined("AAPL").head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)

    """

    @classmethod
    def from_other_refined(
        cls,
        ticker: str,
        /,
        *,
        level: Literal[2, 3, 4] = 2,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from other feature SQL tables.

        Args:
            ticker: Company ticker.
            level: Industry level to aggregate relative features at.
                The industry used according to ``ticker`` is subsampled
                according to this value. Options include:

                    - 2 = major group (e.g., furniture and fixtures)
                    - 3 = industry group (e.g., office furnitures)
                    - 4 = industry (e.g., wood office furniture)

            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Relative annual data dataframe with each tag as a
            separate column. Sorted by filing date.

        Examples:
            >>> finagg.sec.feat.annual.normalized.from_other_refined("AAPL").head(5)  # doctest: +SKIP
                             NORM(LOG_CHANGE(Assets))  NORM(LOG_CHANGE(AssetsCurrent))  NORM(LOG_CHANGE(CommonStockSharesOutstanding)) ...
            fy   filed                                                                                                                 ...
            2010 2010-10-27                 -0.707107                         0.707107                                        0.707107 ...
            2011 2011-10-26                  0.549435                         0.485644                                        0.574215 ...
            2012 2012-10-31                  2.160612                        -0.609935                                        0.802452 ...
            2013 2013-10-30                  0.756833                         0.336215                                       -0.148804 ...
            2014 2014-10-27                  0.244212                         1.384738                                        2.836717 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        company_df = Annual.from_refined(
            ticker, start=start, end=end, engine=engine
        ).reset_index(["filed"])
        filed = company_df["filed"]
        industry_df = IndustryAnnual.from_refined(
            ticker=ticker, level=level, start=start, end=end, engine=engine
        ).reset_index(["filed"])
        company_df = (company_df - industry_df["mean"]) / industry_df["std"]
        company_df["filed"] = filed
        func_cols = utils.get_func_cols(sql.annual)
        company_df[func_cols] = company_df[func_cols].fillna(value=0.0)
        company_df = (
            company_df.reset_index()
            .set_index(["fy", "filed"])
            .rename(lambda x: f"NORM({x})", axis=1)
        )
        company_df = (
            company_df.ffill()
            .dropna()
            .reset_index()
            .drop_duplicates("filed")
            .set_index(["fy", "filed"])
            .sort_index()
        )
        company_df = utils.resolve_col_order(
            sql.normalized_annual, company_df, extra_ignore=["filed"]
        )
        return company_df

    @classmethod
    def from_refined(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from the features SQL table.

        This is the preferred method for accessing features for
        offline analysis (assuming data in the local SQL table
        is current).

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Annual data dataframe with each tag as a
            separate column. Sorted by filing date.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the
                refined SQL table.

        Examples:
            >>> finagg.sec.feat.annual.normalized.from_refined("AAPL").head(5)  # doctest: +SKIP
                             NORM(LOG_CHANGE(Assets))  NORM(LOG_CHANGE(AssetsCurrent))  NORM(LOG_CHANGE(CommonStockSharesOutstanding)) ...
            fy   filed                                                                                                                 ...
            2010 2010-10-27                 -0.707107                         0.707107                                        0.707107 ...
            2011 2011-10-26                  0.549435                         0.485644                                        0.574215 ...
            2012 2012-10-31                  2.160612                        -0.609935                                        0.802452 ...
            2013 2013-10-30                  0.756833                         0.336215                                       -0.148804 ...
            2014 2014-10-27                  0.244212                         1.384738                                        2.836717 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.normalized_annual.name):
            sql.normalized_annual.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.normalized_annual.select()
                    .join(
                        sql.submissions,
                        (sql.submissions.c.cik == sql.normalized_annual.c.cik)
                        & (sql.submissions.c.ticker == ticker),
                    )
                    .where(
                        sql.normalized_annual.c.filed >= start,
                        sql.normalized_annual.c.filed <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(
                f"No industry-normalized annual rows found for {ticker}."
            )
        df = df.drop(columns=["cik"]).set_index(["fy", "filed"]).sort_index()
        return df

    @classmethod
    def get_candidate_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the annual SQL table that MAY BE
        ELIGIBLE to be in the feature's SQL table.

        This is just an alias for :meth:`finagg.sec.feat.Annual.get_ticker_set`.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that may be valid for creating
            industry-normalized annual features that also have at least
            ``lb`` rows for each tag used for constructing the features.

        Examples:
            >>> "AAPL" in finagg.sec.feat.annual.normalized.get_candidate_ticker_set()  # doctest: +SKIP
            True

        """
        return Annual.get_ticker_set(lb=lb, start=start, end=end, engine=engine)

    @classmethod
    def get_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the feature's SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that contain all the columns for creating
            industry-normalized annual features that also have at least
            ``lb`` rows.

        Examples:
            >>> "AAPL" in finagg.sec.feat.annual.normalized.get_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.normalized_annual.name):
            sql.normalized_annual.create(engine)
        with engine.begin() as conn:
            tickers = (
                conn.execute(
                    sa.select(sql.submissions.c.ticker)
                    .join(
                        sql.normalized_annual,
                        sql.normalized_annual.c.cik == sql.submissions.c.cik,
                    )
                    .where(
                        sql.normalized_annual.c.filed >= start,
                        sql.normalized_annual.c.filed <= end,
                    )
                    .group_by(sql.normalized_annual.c.cik)
                    .having(sa.func.count(sql.normalized_annual.c.filed) >= lb)
                )
                .scalars()
                .all()
            )
        return set(tickers)

    @classmethod
    def get_tickers_sorted_by(
        cls,
        column: str,
        /,
        *,
        ascending: bool = True,
        year: int = -1,
        engine: None | Engine = None,
    ) -> list[str]:
        """Get all tickers in the feature's SQL table sorted by a particular
        column.

        Args:
            column: Feature column to sort by.
            ascending: Whether to return results in ascending order according
                to the values in ``column``.
            year: Year to select from. Defaults to the most recent year that
                has data available.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Tickers sorted by a feature column for a particular year.

        Examples:
            >>> ts = finagg.sec.feat.annual.normalized.get_tickers_sorted_by(
            ...         "NORM(EarningsPerShareBasic)",
            ...         year=2020,
            ... )
            >>> "AMD" == ts[0]  # doctest: +SKIP
            True

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.normalized_annual.name):
            sql.normalized_annual.create(engine)
        with engine.begin() as conn:
            if year == -1:
                (max_year,) = conn.execute(
                    sa.select(sa.func.max(sql.normalized_annual.c.fy))
                ).one()
                year = int(max_year)

            tickers = (
                conn.execute(
                    sa.select(sql.submissions.c.ticker)
                    .join(
                        sql.normalized_annual,
                        sql.normalized_annual.c.cik == sql.submissions.c.cik,
                    )
                    .where(
                        sql.normalized_annual.c.fy == year,
                    )
                    .order_by(sql.normalized_annual.c[column])
                )
                .scalars()
                .all()
            )
        if not ascending:
            return list(reversed(tickers))
        return list(tickers)

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from the
        annual SQL tables, transforming them into normalized features, and
        then writing to the refined annual normalized SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the candidate tickers from
                :meth:`NormalizedAnnual.get_candidate_ticker_set`.
            processes: Number of background processes to run in parallel
                when processing ``tickers``.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        tickers = tickers or cls.get_candidate_ticker_set(engine=engine)
        if not tickers:
            logger.info(
                "Skipping finagg.sec.feat.annual.normalized installation because no"
                " tickers were provided or no tickers were found with prerequisite data"
                " (i.e., finagg.sec.feat.annual data)"
            )
            return 0

        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(
            sql.normalized_annual.name
        ):
            sql.normalized_annual.drop(engine, checkfirst=True)
            sql.normalized_annual.create(engine)

        return utils._install(
            cls.from_other_refined,
            cls.to_refined,
            logger,
            list(tickers),
            engine,
            desc="Installing refined SEC industry-normalized annual data",
            processes=processes,
        )

    @classmethod
    def to_refined(
        cls,
        ticker: str,
        df: pd.DataFrame,
        /,
        *,
        engine: None | Engine = None,
    ) -> int:
        """Write the dataframe to the feature store for ``ticker``.

        Args:
            ticker: Company ticker.
            df: Dataframe to store completely as rows in a local SQL
                table.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.normalized_annual.name):
            sql.normalized_annual.create(engine)
        df = df.reset_index(["fy", "filed"])
        df["cik"] = sql.get_cik(ticker, engine=engine)
        with engine.begin() as conn:
            conn.execute(sql.normalized_annual.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df.index)


class Annual:
    """Methods for gathering annual features from SEC EDGAR data.

    The module variable :data:`finagg.sec.feat.annual` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    Examples:
        It doesn't matter which data source you use to gather features.
        They all return equivalent dataframes.

        >>> df1 = finagg.sec.feat.annual.from_api("AAPL").head(5)
        >>> df2 = finagg.sec.feat.annual.from_raw("AAPL").head(5)
        >>> df3 = finagg.sec.feat.annual.from_refined("AAPL").head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
        >>> pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)

    """

    industry = IndustryAnnual()
    """Annual features aggregated for an entire industry.
    The most popular way for accessing the :class:`finagg.sec.feat.IndustryAnnual`
    feature set.

    :meta hide-value:
    """

    normalized = NormalizedAnnual()
    """A company's annual features normalized by its industry.
    The most popular way for accessing the :class:`finagg.sec.feat.NormalizedAnnual`
    feature set.

    :meta hide-value:
    """

    @classmethod
    def _normalize(cls, df: pd.DataFrame, /) -> pd.DataFrame:
        """Normalize annual features columns."""
        df = api.compute_financial_ratios(df)
        df = df.replace([-np.inf, np.inf], np.nan).ffill()
        df = utils.resolve_func_cols(sql.annual, df, drop=True, inplace=True)
        df.columns = df.columns.rename(None)
        df = utils.resolve_col_order(sql.annual, df, extra_ignore=["filed"])
        return df.dropna()

    @classmethod
    def from_api(
        cls, ticker: str, /, *, start: None | str = None, end: None | str = None
    ) -> pd.DataFrame:
        """Get annual features directly from the SEC API.

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.

        Returns:
            Annual data dataframe with each tag as a
            separate column. Sorted by filing date.

        Examples:
            >>> finagg.sec.feat.annual.from_api("AAPL").head(5)  # doctest: +SKIP
                             LOG_CHANGE(Assets)  LOG_CHANGE(AssetsCurrent)  LOG_CHANGE(CommonStockSharesOutstanding) ...
            fy   filed                                                                                               ...
            2010 2010-10-27           -0.089864                  -0.023676                                  0.012840 ...
            2011 2011-10-26            0.272493                   0.278241                                  0.017805 ...
            2012 2012-10-31            0.896033                   0.076422                                  0.014423 ...
            2013 2013-10-30            0.414064                   0.248046                                  0.010630 ...
            2014 2014-10-27            0.161871                   0.239927                                  1.902394 ...

        """
        df = api.company_concept.get_multiple_original(
            api.popular_concepts, ticker=ticker, form="10-K", start=start, end=end
        )
        df = api.group_and_pivot_filings(df, form="10-K")
        return cls._normalize(df)

    @classmethod
    def from_raw(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get annual features from a local SEC SQL table.

        Not all data series are published at the same rate or
        time. Missing rows for less-frequent annual publications
        are forward filled.

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Annual data dataframe with each tag as a
            separate column. Sorted by filing date.

        Examples:
            >>> finagg.sec.feat.annual.from_raw("AAPL").head(5)  # doctest: +SKIP
                             LOG_CHANGE(Assets)  LOG_CHANGE(AssetsCurrent)  LOG_CHANGE(CommonStockSharesOutstanding) ...
            fy   filed                                                                                               ...
            2010 2010-10-27           -0.089864                  -0.023676                                  0.012840 ...
            2011 2011-10-26            0.272493                   0.278241                                  0.017805 ...
            2012 2012-10-31            0.896033                   0.076422                                  0.014423 ...
            2013 2013-10-30            0.414064                   0.248046                                  0.010630 ...
            2014 2014-10-27            0.161871                   0.239927                                  1.902394 ...

        """
        df = _raw.Tags.group_and_pivot_from_raw(
            ticker,
            [concept["tag"] for concept in api.popular_concepts],
            form="10-K",
            start=start,
            end=end,
            engine=engine,
        )
        return cls._normalize(df)

    @classmethod
    def from_refined(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from the refined SQL table.

        This is the preferred method for accessing features for
        offline analysis (assuming data in the local SQL table
        is current).

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Annual data dataframe with each tag as a
            separate column. Sorted by filing date.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the
                refined SQL table.

        Examples:
            >>> finagg.sec.feat.annual.from_refined("AAPL").head(5)  # doctest: +SKIP
                             LOG_CHANGE(Assets)  LOG_CHANGE(AssetsCurrent)  LOG_CHANGE(CommonStockSharesOutstanding) ...
            fy   filed                                                                                               ...
            2010 2010-10-27           -0.089864                  -0.023676                                  0.012840 ...
            2011 2011-10-26            0.272493                   0.278241                                  0.017805 ...
            2012 2012-10-31            0.896033                   0.076422                                  0.014423 ...
            2013 2013-10-30            0.414064                   0.248046                                  0.010630 ...
            2014 2014-10-27            0.161871                   0.239927                                  1.902394 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.annual.name):
            sql.annual.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.annual.select()
                    .join(
                        sql.submissions,
                        (sql.submissions.c.cik == sql.annual.c.cik)
                        & (sql.submissions.c.ticker == ticker),
                    )
                    .where(
                        sql.annual.c.filed >= start,
                        sql.annual.c.filed <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No annual rows found for {ticker}.")
        df = df.drop(columns=["cik"]).set_index(["fy", "filed"]).sort_index()
        return df

    @classmethod
    def get_candidate_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the raw SQL table that MAY BE ELIGIBLE
        to be in the feature's SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that may be valid for creating annual features
            that also have at least ``lb`` rows for each tag used for
            constructing the features.

        Examples:
            >>> "AAPL" in finagg.sec.feat.annual.get_candidate_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.tags.name):
            sql.tags.create(engine)
        with engine.begin() as conn:
            tickers = (
                conn.execute(
                    sa.select(
                        sql.submissions.c.ticker,
                        *[
                            sa.func.sum(
                                sa.case(
                                    {concept["tag"]: 1}, value=sql.tags.c.tag, else_=0
                                )
                            ).label(concept["tag"])
                            for concept in api.popular_concepts
                        ],
                    )
                    .join(sql.tags, sql.tags.c.cik == sql.submissions.c.cik)
                    .where(
                        sql.tags.c.form == "10-K",
                        sql.tags.c.filed >= start,
                        sql.tags.c.filed <= end,
                    )
                    .group_by(sql.tags.c.cik)
                    .having(
                        *[
                            sa.text(f"{concept['tag']} >= {lb}")
                            for concept in api.popular_concepts
                        ]
                    )
                )
                .scalars()
                .all()
            )
        return set(tickers)

    @classmethod
    def get_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the feature's SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that contain all the columns for creating
            annual features that also have at least ``lb`` rows.

        Examples:
            >>> "AAPL" in finagg.sec.feat.annual.get_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.submissions.name):
            sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.annual.name):
            sql.annual.create(engine)
        with engine.begin() as conn:
            tickers = (
                conn.execute(
                    sa.select(sql.submissions.c.ticker)
                    .join(sql.annual, sql.annual.c.cik == sql.submissions.c.cik)
                    .where(
                        sql.annual.c.filed >= start,
                        sql.annual.c.filed <= end,
                    )
                    .group_by(sql.annual.c.cik)
                    .having(sa.func.count(sql.annual.c.filed) >= lb)
                )
                .scalars()
                .all()
            )
        return set(tickers)

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from the
        raw SQL tables, transforming them into annual features, and then
        writing to the refined annual SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the candidate tickers from
                :meth:`Annual.get_candidate_ticker_set`.
            processes: Number of background processes to run in parallel
                when processing ``tickers``.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        tickers = tickers or cls.get_candidate_ticker_set(engine=engine)
        if not tickers:
            logger.info(
                "Skipping finagg.sec.feat.annual installation because no tickers were"
                " provided or no tickers were found with prerequisite data (i.e.,"
                " finagg.sec.feat.submissions and finagg.sec.feat.tags data)"
            )
            return 0

        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.annual.name):
            sql.annual.drop(engine, checkfirst=True)
            sql.annual.create(engine)

        return utils._install(
            cls.from_raw,
            cls.to_refined,
            logger,
            list(tickers),
            engine,
            desc="Installing refined SEC annual data",
            processes=processes,
        )

    @classmethod
    def to_refined(
        cls,
        ticker: str,
        df: pd.DataFrame,
        /,
        *,
        engine: None | Engine = None,
    ) -> int:
        """Write the given dataframe to the refined feature table
        while using the ticker ``ticker``.

        Args:
            ticker: Company ticker.
            df: Dataframe to store as rows in a local SQL table
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.annual.name):
            sql.annual.create(engine)
        df = df.reset_index(["fy", "filed"])
        df["cik"] = sql.get_cik(ticker, engine=engine)
        with engine.begin() as conn:
            conn.execute(sql.annual.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df.index)



"""FRED SQLAlchemy interfaces."""


import sqlalchemy as sa

metadata = sa.MetaData()
"""The metadata associated with all SQL tables defined in this module.

:meta hide-value:
"""

series = sa.Table(
    "fred.raw.series",
    metadata,
    sa.Column("series_id", sa.String, primary_key=True, doc="Economic series ID."),
    sa.Column(
        "realtime_start",
        sa.String,
        primary_key=True,
        doc="Start date for values according to their publication date.",
    ),
    sa.Column(
        "realtime_end",
        sa.String,
        primary_key=True,
        doc="End date for values according to their publication date.",
    ),
    sa.Column(
        "date", sa.String, primary_key=True, doc="Series value publication date."
    ),
    sa.Column(
        "value",
        sa.Float,
        doc="Economic series value for a particular date.",
    ),
)
"""SQL table for storing raw data as managed by
:data:`finagg.fred.feat.series` (an alias for
:class:`finagg.fred.feat.Series`).

:meta hide-value:
"""

economic = sa.Table(
    "fred.refined.economic",
    metadata,
    sa.Column(
        "date",
        sa.String,
        primary_key=True,
        doc="Economic data series release date.",
    ),
    sa.Column(
        "CIVPART", sa.Float, nullable=False, doc="Labor force participation rate."
    ),
    sa.Column(
        "LOG_CHANGE(CPIAUCNS)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in consumer price index between days.",
    ),
    sa.Column(
        "LOG_CHANGE(CSUSHPINSA)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in S&P/Case-Shiller national home price index between"
            " days."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(DJIA)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in Dow Jones Industrial index.",
    ),
    sa.Column("FEDFUNDS", sa.Float, nullable=False, doc="Federal funds interest rate."),
    sa.Column(
        "LOG_CHANGE(GDP)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in gross domestic product between days.",
    ),
    sa.Column(
        "LOG_CHANGE(GDPC1)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in real gross domestic product between days.",
    ),
    sa.Column("GS10", sa.Float, nullable=False, doc="10-Year treasury yield."),
    sa.Column(
        "LOG_CHANGE(M2)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in money stock measures (i.e., savings and related"
            " balances) between days."
        ),
    ),
    sa.Column(
        "MICH",
        sa.Float,
        nullable=False,
        doc="University of Michigan: inflation expectation.",
    ),
    sa.Column(
        "LOG_CHANGE(NASDAQ100)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in Nasdaq 100 index.",
    ),
    sa.Column(
        "LOG_CHANGE(NASDAQCOM)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in Nasdaq Composite index.",
    ),
    sa.Column("PSAVERT", sa.Float, nullable=False, doc="Personal savings rate."),
    sa.Column(
        "LOG_CHANGE(SP500)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in S&P 500 index.",
    ),
    sa.Column(
        "UMCSENT",
        sa.Float,
        nullable=False,
        doc="University of Michigan: consumer sentiment.",
    ),
    sa.Column("UNRATE", sa.Float, nullable=False, doc="Unemployment rate."),
    sa.Column(
        "LOG_CHANGE(WALCL)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in US assets, total assets (less eliminations from"
            " consolidation) between days."
        ),
    ),
)
"""SQL table for storing refined data as managed by :data:`finagg.fred.feat.economic`
(an alias for :class:`finagg.fred.feat.Economic`).

:meta hide-value:
"""



"""CLI and tools for aggregating tickers in common indices."""

import logging

import click

from .. import backend
from . import api as _api
from . import sql as _sql

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


@click.group(help="Tools for managing popular indices's ticker data.")
def entry_point() -> None:
    ...


@entry_point.command(
    help=(
        "Drop and recreate tables, and install popular indices's ticker data "
        "into the SQL database."
    ),
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    default=False,
    help="Whether to install all defined tables.",
)
@click.option(
    "--djia",
    is_flag=True,
    default=False,
    help="Whether to install the DJIA ticker table.",
)
@click.option(
    "--djia",
    is_flag=True,
    default=False,
    help="Whether to install the DJIA ticker table.",
)
@click.option(
    "--sp500",
    is_flag=True,
    default=False,
    help="Whether to install the S&P 500 ticker table.",
)
@click.option(
    "--nasdaq100",
    is_flag=True,
    default=False,
    help="Whether to install the Nasdaq 100 ticker table.",
)
def install(
    all_: bool = False,
    djia: bool = False,
    sp500: bool = False,
    nasdaq100: bool = False,
) -> int:
    total_rows = 0
    with backend.engine.begin() as conn:
        if all_ or djia:
            _sql.djia.drop(conn, checkfirst=True)
            _sql.djia.create(conn)

            df = _api.djia.get()
            rowcount = len(df.index)
            conn.execute(
                _sql.djia.insert(), df.to_dict(orient="records")  # type: ignore[arg-type]
            )
            logger.info(f"Inserted {rowcount} rows into the DJIA table")
            total_rows += rowcount

        if all_ or sp500:
            _sql.sp500.drop(conn, checkfirst=True)
            _sql.sp500.create(conn)

            df = _api.sp500.get()
            rowcount = len(df.index)
            conn.execute(
                _sql.sp500.insert(), df.to_dict(orient="records")  # type: ignore[arg-type]
            )
            logger.info(f"Inserted {rowcount} rows into the S&P 500 table")
            total_rows += rowcount

        if all_ or nasdaq100:
            _sql.nasdaq100.drop(conn, checkfirst=True)
            _sql.nasdaq100.create(conn)

            df = _api.nasdaq100.get()
            rowcount = len(df.index)
            conn.execute(
                _sql.nasdaq100.insert(), df.to_dict(orient="records")  # type: ignore[arg-type]
            )
            logger.info(f"Inserted {rowcount} rows into the Nasdaq 100 table")
            total_rows += rowcount

    if all_ or djia or sp500 or nasdaq100:
        logger.info(f"{total_rows} total rows inserted for {__package__}")
    else:
        logger.info(
            f"Skipping {__package__} installation because no installation "
            "options are provided"
        )

    return total_rows



import finagg


def test_djia_get() -> None:
    df = finagg.indices.api.djia.get()
    assert "AAPL" in set(df["ticker"])


def test_get_ticker_set() -> None:
    tickers = finagg.indices.api.get_ticker_set()
    assert "AAPL" in tickers


def test_nasdaq100_get() -> None:
    df = finagg.indices.api.nasdaq100.get()
    assert "AAPL" in set(df["ticker"])


def test_sp500_get() -> None:
    df = finagg.indices.api.sp500.get()
    assert "AAPL" in set(df["ticker"])



"""CLI and tools for yfinance."""

import logging
import multiprocessing as mp
from typing import Literal

import click

from .. import indices, sec, utils
from . import feat as _feat

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


@click.group(help="Yahoo! Finance tools.")
def entry_point() -> None:
    ...


@entry_point.command(
    help="Install Yahoo! Finance stock price and price feature data.",
)
@click.option(
    "--raw",
    type=click.Choice(["prices"]),
    multiple=True,
    help=(
        "Raw tables to install. `prices` indicates daily historical stock "
        "price data. `prices` must be specified to enable installing refined "
        "data using the `refined` flag."
    ),
)
@click.option(
    "--refined",
    type=click.Choice(["daily"]),
    multiple=True,
    help=(
        "Refined tables to install. This requires raw data to be "
        "installed beforehand using the `--raw` flag or for the "
        "`--raw` flag to be set when this option is provided."
    ),
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    default=False,
    help="Whether to install all defined tables (including all refined tables).",
)
@click.option(
    "--ticker",
    "-t",
    multiple=True,
    help=(
        "Ticker whose data is attempted to be downloaded and inserted into "
        "the SQL tables. Multiple tickers can be specified by providing "
        "multiple `ticker` options, by separating tickers with a comma (e.g., "
        "`AAPL,MSFT,NVDA`), or by providing tickers in a CSV file by "
        "specifying a file path (e.g., `bank_tickers.txt`). The CSV file "
        "can be formatted such that there's one ticker per line or multiple "
        "tickers per line (delimited by a comma). The tickers specified "
        "by this option are combined with the tickers specified by the "
        "`ticker-set` option."
    ),
)
@click.option(
    "--ticker-set",
    "-ts",
    "ticker_set",
    type=click.Choice(["indices", "sec"]),
    default=None,
    help=(
        "Set of tickers whose data is attempted to be downloaded and "
        "inserted into the SQL tables. 'indices' indicates the set "
        "of tickers from the three most popular indices (DJIA, "
        "Nasdaq 100, and S&P 500). 'sec' indicates all the tickers that "
        "have data available through the SEC API (which is approximately "
        "all publicly-traded US companies). 'indices' will effectively "
        "only attempt to download and install data for relatively "
        "popular and large market cap companies, while 'sec' will "
        "attempt to download and install data for nearly all "
        "publicly-traded US companies. Choosing 'indices' will be fast, "
        "while choosing 'sec' will be slow but will include more diverse data."
    ),
)
@click.option(
    "--processes",
    "-n",
    type=int,
    default=mp.cpu_count() - 1,
    help=(
        "Number of backgruond processes to run in parallel when installing data. Note,"
        " not all tables support installations with multiprocessing."
    ),
)
@click.option(
    "--recreate-tables",
    "-r",
    is_flag=True,
    default=False,
    help=(
        "Whether to reset the tables associated with the install options by "
        "dropping and recreating them."
    ),
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Sets the log level to DEBUG to show installation errors for each ticker.",
)
def install(
    raw: list[Literal["prices"]] = [],
    refined: list[Literal["daily"]] = [],
    all_: bool = False,
    ticker: list[str] = [],
    ticker_set: None | Literal["indices", "sec"] = None,
    processes: int = mp.cpu_count() - 1,
    recreate_tables: bool = False,
    verbose: bool = False,
) -> int:
    if verbose:
        logging.getLogger(__package__).setLevel(logging.DEBUG)

    total_rows = 0
    all_raw = set()
    if all_:
        all_raw = {"prices"}
    elif raw:
        all_raw = set(raw)

    all_tickers = utils.expand_csv(ticker)
    if all_raw:
        match ticker_set:
            case "indices":
                all_tickers |= indices.api.get_ticker_set()
            case "sec":
                all_tickers |= sec.api.get_ticker_set()

        if not all_tickers:
            logger.info(
                f"Skipping {__package__} installation because no tickers were "
                "provided (by the `ticker` option or by the `ticker-set` option)"
            )
            return total_rows

        if "prices" in all_raw:
            total_rows += _feat.prices.install(
                all_tickers, processes=processes, recreate_tables=recreate_tables
            )

    all_refined = set()
    if all_:
        all_refined = {"daily"}
    elif refined:
        all_refined = set(refined)

    if "daily" in all_refined:
        total_rows += _feat.daily.install(
            tickers=all_tickers, processes=processes, recreate_tables=recreate_tables
        )

    if all_ or all_refined or all_raw:
        if total_rows:
            logger.info(f"{total_rows} total rows inserted for {__package__}")
        else:
            logger.warning(
                f"No rows were inserted for {__package__}. This could be an error if"
                " installations were skipped. Set the verbose flag with the"
                " `--verbose/-v` option to enable debug logging."
            )
    else:
        logger.info(
            f"Skipping {__package__} installation because no installation "
            "options are provided"
        )
    return total_rows


@entry_point.command(
    help="Update Yahoo! Finance stock price and price feature data.",
)
@click.option(
    "--raw",
    type=click.Choice(["prices"]),
    multiple=True,
    help=(
        "Raw tables to update. `prices` indicates daily historical stock "
        "price data. `prices` must be specified to enable updating refined "
        "data using the `refined` flag."
    ),
)
@click.option(
    "--refined",
    type=click.Choice(["daily"]),
    multiple=True,
    help=(
        "Refined tables to update. This requires raw data to be "
        "updated beforehand using the `--raw` flag or for the "
        "`--raw` flag to be set when this option is provided."
    ),
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    default=False,
    help="Whether to update all defined tables (including all refined tables).",
)
@click.option(
    "--ticker",
    "-t",
    multiple=True,
    help=(
        "Ticker whose data is attempted to be downloaded and inserted into "
        "the SQL tables. Multiple tickers can be specified by providing "
        "multiple `ticker` options, by separating tickers with a comma (e.g., "
        "`AAPL,MSFT,NVDA`), or by providing tickers in a CSV file by "
        "specifying a file path (e.g., `bank_tickers.txt`). The CSV file "
        "can be formatted such that there's one ticker per line or multiple "
        "tickers per line (delimited by a comma)."
    ),
)
@click.option(
    "--processes",
    "-n",
    type=int,
    default=mp.cpu_count() - 1,
    help=(
        "Number of backgruond processes to run in parallel when updating. Note,"
        " not all tables support updates with multiprocessing."
    ),
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Sets the log level to DEBUG to show update errors for each ticker.",
)
def update(
    raw: list[Literal["prices"]] = [],
    refined: list[Literal["daily"]] = [],
    all_: bool = False,
    ticker: list[str] = [],
    processes: int = mp.cpu_count() - 1,
    verbose: bool = False,
) -> int:
    if verbose:
        logging.getLogger(__package__).setLevel(logging.DEBUG)

    total_rows = 0
    all_raw = set()
    if all_:
        all_raw = {"prices"}
    elif raw:
        all_raw = set(raw)

    all_tickers = utils.expand_csv(ticker)
    if all_raw:
        if "prices" in all_raw:
            total_rows += _feat.prices.update(
                all_tickers,
                processes=processes,
            )

    all_refined = set()
    if all_:
        all_refined = {"daily"}
    elif refined:
        all_refined = set(refined)

    if "daily" in all_refined:
        total_rows += _feat.daily.update(
            tickers=all_tickers,
            processes=processes,
        )

    if all_ or all_refined or all_raw:
        if total_rows:
            logger.info(f"{total_rows} total rows inserted for {__package__}")
        else:
            logger.warning(
                f"No rows were inserted for {__package__}. This could be an error if"
                " updates were skipped. Set the verbose flag with the"
                " `--verbose/-v` option to enable debug logging."
            )
    else:
        logger.info(
            f"Skipping {__package__} update because no update options are provided"
        )
    return total_rows



"""Refined FRED features (features aggregated from raw tables)."""


import logging

import pandas as pd
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound

from ... import backend, utils
from .. import api, sql

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class Economic:
    """Methods for gathering economic data series from FRED sources.

    The module variable :data:`finagg.fred.feat.economic` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    Examples:
        It doesn't matter which data source you use to gather features.
        They all return equivalent dataframes.

        >>> df1 = finagg.fred.feat.economic.from_api().head(5)
        >>> df2 = finagg.fred.feat.economic.from_raw().head(5)
        >>> df3 = finagg.fred.feat.economic.from_refined().head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
        >>> pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)

    """

    @classmethod
    def _normalize(cls, df: pd.DataFrame, /) -> pd.DataFrame:
        """Normalize economic features columns."""
        df = (
            df.pivot(index="date", values="value", columns="series_id")
            .sort_index()
            .astype(float)
            .ffill()
            .dropna()
        )
        df = utils.resolve_func_cols(sql.economic, df, drop=True, inplace=True)
        df.columns = df.columns.rename(None)
        df = utils.resolve_col_order(sql.economic, df)
        return df.dropna()

    @classmethod
    def from_api(
        cls, *, start: None | str = None, end: None | str = None
    ) -> pd.DataFrame:
        """Get economic features directly from the FRED API.

        Not all data series are published at the same rate or
        time. Missing rows for less-frequent economic series
        are forward filled.

        Args:
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.

        Returns:
            Economic data series dataframe with each series
            as a separate column. Sorted by date.

        Examples:
            >>> finagg.fred.feat.economic.from_api().head(5)  # doctest: +SKIP
                        CIVPART  LOG_CHANGE(CPIAUCNS)  LOG_CHANGE(CSUSHPINSA)  FEDFUNDS ...
            date                                                                        ...
            2014-10-06     62.8                   0.0                     0.0      0.09 ...
            2014-10-08     62.8                   0.0                     0.0      0.09 ...
            2014-10-13     62.8                   0.0                     0.0      0.09 ...
            2014-10-15     62.8                   0.0                     0.0      0.09 ...
            2014-10-20     62.8                   0.0                     0.0      0.09 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        dfs = []
        for series_id in api.popular_series:
            df = api.series.observations.get_original_observations(
                series_id,
                observation_start=start,
                observation_end=end,
            )
            dfs.append(df)
        df = pd.concat(dfs)
        return cls._normalize(df)

    @classmethod
    def from_raw(
        cls,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get economic features from local FRED SQL tables.

        Not all data series are published at the same rate or
        time. Missing rows for less-frequent economic series
        are forward filled.

        Args:
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Economic data series dataframe with each series
            as a separate column. Sorted by date.

        Raises:
            `NoResultFound`: If there are no rows in the raw SQL table.

        Examples:
            >>> finagg.fred.feat.economic.from_raw().head(5)  # doctest: +SKIP
                        CIVPART  LOG_CHANGE(CPIAUCNS)  LOG_CHANGE(CSUSHPINSA)  FEDFUNDS ...
            date                                                                        ...
            2014-10-06     62.8                   0.0                     0.0      0.09 ...
            2014-10-08     62.8                   0.0                     0.0      0.09 ...
            2014-10-13     62.8                   0.0                     0.0      0.09 ...
            2014-10-15     62.8                   0.0                     0.0      0.09 ...
            2014-10-20     62.8                   0.0                     0.0      0.09 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.series.name):
            sql.series.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.series.select().where(
                        sql.series.c.series_id.in_(api.popular_series),
                        sql.series.c.date >= start,
                        sql.series.c.date <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No economic rows found.")
        return cls._normalize(df)

    @classmethod
    def from_refined(
        cls,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from the feature-dedicated local SQL tables.

        This is the preferred method for accessing features for
        offline analysis (assuming data in the local SQL tables
        is current).

        Args:
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Economic data series dataframe with each series
            as a separate column. Sorted by date.

        Raises:
            `NoResultFound`: If there are no rows in the refined SQL table.

        Examples:
            >>> finagg.fred.feat.economic.from_refined().head(5)  # doctest: +SKIP
                        CIVPART  LOG_CHANGE(CPIAUCNS)  LOG_CHANGE(CSUSHPINSA)  FEDFUNDS ...
            date                                                                        ...
            2014-10-06     62.8                   0.0                     0.0      0.09 ...
            2014-10-08     62.8                   0.0                     0.0      0.09 ...
            2014-10-13     62.8                   0.0                     0.0      0.09 ...
            2014-10-15     62.8                   0.0                     0.0      0.09 ...
            2014-10-20     62.8                   0.0                     0.0      0.09 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.economic.name):
            sql.economic.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.economic.select().where(
                        sql.economic.c.date >= start, sql.economic.c.date <= end
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No economic rows found.")
        df = df.set_index("date").sort_index()
        return df

    @classmethod
    def install(
        cls,
        *,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install economic data by pulling data from the raw SQL tables,
        transforming them into economic features, and then writing to the
        refined economic data SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.economic.name):
            sql.economic.drop(engine, checkfirst=True)
            sql.economic.create(engine)

        total_rows = 0
        try:
            df = cls.from_raw(engine=engine)
            rowcount = len(df.index)
            if rowcount:
                cls.to_refined(df, engine=engine)
                total_rows += rowcount
                logger.debug(f"{rowcount} economic feature rows inserted")
            else:
                logger.debug("Skipping economic features due to missing data")
        except Exception as e:
            logger.debug(f"Skipping economic features", exc_info=e)
        return total_rows

    @classmethod
    def to_refined(
        cls,
        df: pd.DataFrame,
        /,
        *,
        engine: None | Engine = None,
    ) -> int:
        """Write the dataframe to the feature store.

        Args:
            df: Dataframe to store completely as rows in a local SQL
                table.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.economic.name):
            sql.economic.create(engine)
        df = df.reset_index("date")
        with engine.begin() as conn:
            conn.execute(sql.economic.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df.index)



import pytest

from finagg.portfolio import Portfolio, Position

CHANGE_MULTIPLE = 0.5
INITIAL_CASH = 10_000
INITIAL_COST = 100
INITIAL_QUANTITY = 10


@pytest.fixture
def portfolio() -> Portfolio:
    return Portfolio(INITIAL_CASH)


@pytest.fixture
def position() -> Position:
    return Position(INITIAL_COST, INITIAL_QUANTITY)


def test_portfolio_buy(portfolio: Portfolio, position: Position) -> None:
    portfolio.buy("TEST", INITIAL_COST, INITIAL_QUANTITY)
    assert portfolio.cash == (INITIAL_CASH - (INITIAL_COST * INITIAL_QUANTITY))
    assert portfolio["TEST"] == position
    assert portfolio["TEST"].total_cost_basis == position.total_cost_basis
    assert portfolio["TEST"].quantity == position.quantity


def test_portfolio_sell(portfolio: Portfolio) -> None:
    portfolio.buy("TEST", INITIAL_COST, INITIAL_QUANTITY)
    portfolio.sell("TEST", CHANGE_MULTIPLE * INITIAL_COST, INITIAL_QUANTITY)
    assert portfolio.cash == (
        INITIAL_CASH - (CHANGE_MULTIPLE * INITIAL_COST * INITIAL_QUANTITY)
    )
    assert "TEST" not in portfolio


def test_portfolio_total_dollar_change(portfolio: Portfolio) -> None:
    portfolio.buy("TEST", INITIAL_CASH, 1)
    assert portfolio.total_dollar_change({"TEST": CHANGE_MULTIPLE * INITIAL_CASH}) == (
        -CHANGE_MULTIPLE * INITIAL_CASH
    )
    assert portfolio.total_dollar_change({"TEST": 0.0}) == -INITIAL_CASH


def test_portfolio_total_percent_change(portfolio: Portfolio) -> None:
    portfolio.buy("TEST", INITIAL_CASH, 1)
    assert (
        portfolio.total_percent_change({"TEST": CHANGE_MULTIPLE * INITIAL_CASH}) == -0.5
    )
    assert portfolio.total_percent_change({"TEST": 0.0}) == -1


def test_position_buy(position: Position) -> None:
    assert position.average_cost_basis == INITIAL_COST
    assert position.total_cost_basis == (INITIAL_COST * INITIAL_QUANTITY)
    position.buy(CHANGE_MULTIPLE * INITIAL_COST, INITIAL_QUANTITY)
    assert position.average_cost_basis == (
        INITIAL_COST - INITIAL_COST * (CHANGE_MULTIPLE / 2)
    )
    assert position.total_cost_basis == (
        (1 + CHANGE_MULTIPLE) * INITIAL_COST * INITIAL_QUANTITY
    )


def test_position_sell(position: Position) -> None:
    assert position.average_cost_basis == INITIAL_COST
    assert position.total_cost_basis == (INITIAL_COST * INITIAL_QUANTITY)
    position.sell(
        CHANGE_MULTIPLE * INITIAL_COST,
        CHANGE_MULTIPLE * INITIAL_QUANTITY,
    )
    assert position.average_cost_basis == INITIAL_COST
    assert position.total_cost_basis == (
        (1 - CHANGE_MULTIPLE) * INITIAL_COST * INITIAL_QUANTITY
    )


def test_position_total_dollar_change(position: Position) -> None:
    assert position.average_cost_basis == INITIAL_COST
    assert position.total_cost_basis == (INITIAL_COST * INITIAL_QUANTITY)
    assert (
        position.total_dollar_change(CHANGE_MULTIPLE * INITIAL_COST)
        == -CHANGE_MULTIPLE * position.average_cost_basis * position.quantity
    )


def test_position_total_percent_change(position: Position) -> None:
    assert position.average_cost_basis == INITIAL_COST
    assert position.total_cost_basis == (INITIAL_COST * INITIAL_QUANTITY)
    assert (
        position.total_percent_change(CHANGE_MULTIPLE * INITIAL_COST)
        == -CHANGE_MULTIPLE
    )



"""Testing utils used for ``finagg``'s own unit tests."""

import pathlib
from typing import Generator

import sqlalchemy as sa
from sqlalchemy.engine import Engine


def sqlite_engine(
    path: str,
    /,
    *,
    metadata: None | sa.MetaData = None,
    table: None | sa.Table = None,
) -> Generator[Engine, None, None]:
    """Yield a test database engine that's cleaned-up after
    usage.

    Args:
        path: Path to SQLite database file.
        metadata: Optional metadata for creating and dropping
            tables before and after yielding the engine,
            respectively.
        table: Optional table for creating and dropping before
            and after yielding the engine, respectively.

    Returns:
        A database engine that's subsequently disposed of
        and whose respective database file is deleted
        after use.

    Raises:
        `ValueError`: If both ``metadata`` and ``table`` are provided.

    Examples:
        Using the testing util as a pytest fixture.

        >>> import pytest
        >>> from sqlalchemy.engine import Engine
        >>> @pytest.fixture
        ... def engine() -> Engine:
        ...     yield from finagg.testing.sqlite_engine("/path/to/db.sqlite")

    """
    if metadata and table:
        raise ValueError("`metadata` and `table` are mutally exclusive")

    path_obj = pathlib.Path(path)
    path_obj = path_obj.with_stem(f"{path_obj.stem}_test")
    url = f"sqlite:///{path_obj}"
    engine = sa.create_engine(url)
    if metadata is not None:
        metadata.create_all(engine)
    if table is not None:
        table.create(engine)
    yield engine
    if metadata is not None:
        metadata.drop_all(engine)
    if table is not None:
        table.drop(engine)
    engine.dispose()
    path_obj.unlink()



"""SEC SQLAlchemy interfaces."""

from typing import Any, Literal

import sqlalchemy as sa
from sqlalchemy.engine import Engine

from .. import backend

metadata = sa.MetaData()
"""The metadata associated with all SQL tables defined in this module.

:meta hide-value:
"""

submissions = sa.Table(
    "sec.raw.submissions",
    metadata,
    sa.Column("cik", sa.String, primary_key=True, doc="Unique SEC ID."),
    sa.Column("ticker", sa.String, nullable=False, doc="Company ticker."),
    sa.Column(
        "entity_type", sa.String, doc="Type of company standing (e.g., operating)."
    ),
    sa.Column("sic", sa.String, nullable=False, doc="Industry code."),
    sa.Column("sic_description", sa.String, doc="Industry code description."),
    sa.Column("name", sa.String, doc="Company name."),
    sa.Column(
        "exchanges",
        sa.String,
        doc="Comma-separated exchanges the company is found on.",
    ),
    sa.Column("ein", sa.String, doc="Entity identification number."),
    sa.Column("description", sa.String, doc="Entity description (often empty/null)."),
    sa.Column("category", sa.String, doc="SEC entity category."),
    sa.Column(
        "fiscal_year_end",
        sa.String,
        doc="The company's last day of the fiscal year (MMDD).",
    ),
)
"""SQL table for storing raw data as managed by
:data:`finagg.sec.feat.submissions` (an alias for
:class:`finagg.sec.feat.Submissions`).

:meta hide-value:
"""

tags = sa.Table(
    "sec.raw.tags",
    metadata,
    sa.Column(
        "cik",
        sa.String,
        sa.ForeignKey(submissions.c.cik, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique SEC ID.",
    ),
    sa.Column("accn", sa.String, doc="Unique submission/access number."),
    sa.Column(
        "taxonomy", sa.String, doc="XBRL taxonomy the submission's tag belongs to."
    ),
    sa.Column(
        "tag",
        sa.String,
        primary_key=True,
        doc="XBRL submission tag (e.g., NetIncomeLoss).",
    ),
    sa.Column(
        "form", sa.String, nullable=False, doc="Submission form type (e.g., 10-Q)."
    ),
    sa.Column(
        "units",
        sa.String,
        nullable=False,
        doc="Unit of measurements for tag value (e.g., USD or shares).",
    ),
    sa.Column(
        "fy", sa.Integer, primary_key=True, doc="Fiscal year the submission is for."
    ),
    sa.Column(
        "fp",
        sa.String,
        primary_key=True,
        doc="Fiscal period the submission is for (e.g., Q1 or FY).",
    ),
    sa.Column(
        "start",
        sa.String,
        nullable=True,
        doc="When the tag's value's measurements started.",
    ),
    sa.Column(
        "end",
        sa.String,
        nullable=True,
        doc="When the tag's value's measurements ended.",
    ),
    sa.Column(
        "filed",
        sa.String,
        nullable=False,
        doc="When the submission was actually filed.",
    ),
    sa.Column(
        "frame",
        sa.String,
        nullable=True,
        doc="Often a concatenation of `fy` and `fp`.",
    ),
    sa.Column(
        "label",
        sa.String,
        nullable=True,
        doc="More human readable version of `tag`.",
    ),
    sa.Column(
        "description",
        sa.String,
        nullable=True,
        doc="Long description of `tag` and `label`.",
    ),
    sa.Column("entity", sa.String, doc="Company name."),
    sa.Column("value", sa.Float, nullable=False, doc="Tag value with units `units`."),
)
"""SQL table for storing raw data as managed by :data:`finagg.sec.feat.tags`
(an alias for :class:`finagg.sec.feat.Tags`).

:meta hide-value:
"""

annual = sa.Table(
    "sec.refined.annual",
    metadata,
    sa.Column(
        "cik",
        sa.String,
        sa.ForeignKey(submissions.c.cik, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique company ticker.",
    ),
    sa.Column("filed", sa.String, nullable=False, doc="Filing date."),
    sa.Column("fy", sa.Integer, primary_key=True, doc="Fiscal year the value is for."),
    sa.Column(
        "LOG_CHANGE(Assets)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's total assets between years.",
    ),
    sa.Column(
        "LOG_CHANGE(AssetsCurrent)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's current assets between years.",
    ),
    sa.Column(
        "LOG_CHANGE(CommonStockSharesOutstanding)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's stock shares outstanding between years.",
    ),
    sa.Column(
        "LOG_CHANGE(InventoryNet)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's inventory between years.",
    ),
    sa.Column(
        "LOG_CHANGE(Liabilities)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's total liabilities between years.",
    ),
    sa.Column(
        "LOG_CHANGE(LiabilitiesCurrent)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's current liabilities between years.",
    ),
    sa.Column(
        "LOG_CHANGE(StockholdersEquity)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's stockholder's equity between years.",
    ),
    sa.Column(
        "AssetCoverageRatio",
        sa.Float,
        nullable=False,
        doc="Total assets minus short term liabilities over total liabilities.",
    ),
    sa.Column(
        "BookRatio",
        sa.Float,
        nullable=False,
        doc="Total assets minus total liabilities over outstanding shares.",
    ),
    sa.Column(
        "DebtEquityRatio",
        sa.Float,
        nullable=False,
        doc="Total liabilities over stock holder's equity.",
    ),
    sa.Column(
        "EarningsPerShareBasic", sa.Float, nullable=False, doc="Earnings per share."
    ),
    sa.Column(
        "QuickRatio",
        sa.Float,
        nullable=False,
        doc="Current assets minus inventory over current liabilities.",
    ),
    sa.Column(
        "ReturnOnAssets",
        sa.Float,
        nullable=False,
        doc="Net income/loss over total assets.",
    ),
    sa.Column(
        "ReturnOnEquity",
        sa.Float,
        nullable=False,
        doc="Net income/loss over stockholder's equity.",
    ),
    sa.Column(
        "WorkingCapitalRatio",
        sa.Float,
        nullable=False,
        doc="Current assets over current liabilities.",
    ),
)
"""SQL table for storing refined data as managed by :data:`finagg.sec.feat.annual`
(an alias for :class:`finagg.sec.feat.Annual`).

:meta hide-value:
"""

normalized_annual = sa.Table(
    "sec.refined.annual.normalized",
    metadata,
    sa.Column(
        "cik",
        sa.String,
        sa.ForeignKey(submissions.c.cik, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique company ticker.",
    ),
    sa.Column("filed", sa.String, nullable=False, doc="Filing date."),
    sa.Column("fy", sa.Integer, primary_key=True, doc="Fiscal year the value is for."),
    sa.Column(
        "NORM(LOG_CHANGE(Assets))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's total assets between years "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(AssetsCurrent))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's current assets between years "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(CommonStockSharesOutstanding))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's stock shares outstanding between years "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(InventoryNet))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's inventory between years "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(Liabilities))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's total liabilities between years "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(LiabilitiesCurrent))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's current liabilities between years "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(StockholdersEquity))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's stockholder's equity between "
            "years normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(AssetCoverageRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Total assets minus short term liabilities over total liabilities "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(BookRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Total assets minus total liabilities over outstanding shares "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(DebtEquityRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Total liabilities over stock holder's equity normalized against "
            "the company's industry."
        ),
    ),
    sa.Column(
        "NORM(EarningsPerShareBasic)",
        sa.Float,
        nullable=False,
        doc="Earnings per share normalized against the company's industry.",
    ),
    sa.Column(
        "NORM(QuickRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Current assets minus inventory over current liabilities normalized "
            "against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(ReturnOnAssets)",
        sa.Float,
        nullable=False,
        doc=(
            "Net income/loss over total assets normalized against the company's"
            " industry."
        ),
    ),
    sa.Column(
        "NORM(ReturnOnEquity)",
        sa.Float,
        nullable=False,
        doc=(
            "Net income/loss over stockholder's equity normalized against the company's"
            " industry."
        ),
    ),
    sa.Column(
        "NORM(WorkingCapitalRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Current assets over current liabilities normalized against the company's"
            " industry."
        ),
    ),
)
"""SQL table for storing refined data as managed by
:attr:`finagg.sec.feat.Annual.normalized` (an alias for
:class:`finagg.sec.feat.NormalizedAnnual`).

:meta hide-value:
"""

quarterly = sa.Table(
    "sec.refined.quarterly",
    metadata,
    sa.Column(
        "cik",
        sa.String,
        sa.ForeignKey(submissions.c.cik, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique company ticker.",
    ),
    sa.Column("filed", sa.String, nullable=False, doc="Filing date."),
    sa.Column("fy", sa.Integer, primary_key=True, doc="Fiscal year the value is for."),
    sa.Column(
        "fp",
        sa.String,
        primary_key=True,
        doc="Fiscal period the value is for (e.g., Q1 or FY).",
    ),
    sa.Column(
        "LOG_CHANGE(Assets)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's total assets between quarters.",
    ),
    sa.Column(
        "LOG_CHANGE(AssetsCurrent)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's current assets between quarters.",
    ),
    sa.Column(
        "LOG_CHANGE(CommonStockSharesOutstanding)",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's stock shares outstanding between"
            " quarters."
        ),
    ),
    sa.Column(
        "LOG_CHANGE(InventoryNet)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's inventory between quarters.",
    ),
    sa.Column(
        "LOG_CHANGE(Liabilities)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's total liabilities between quarters.",
    ),
    sa.Column(
        "LOG_CHANGE(LiabilitiesCurrent)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's current liabilities between quarters.",
    ),
    sa.Column(
        "LOG_CHANGE(StockholdersEquity)",
        sa.Float,
        nullable=False,
        doc="Logarithmic change in a company's stockholder's equity between quarters.",
    ),
    sa.Column(
        "AssetCoverageRatio",
        sa.Float,
        nullable=False,
        doc="Total assets minus short term liabilities over total liabilities.",
    ),
    sa.Column(
        "BookRatio",
        sa.Float,
        nullable=False,
        doc="Total assets minus total liabilities over outstanding shares.",
    ),
    sa.Column(
        "DebtEquityRatio",
        sa.Float,
        nullable=False,
        doc="Total liabilities over stock holder's equity.",
    ),
    sa.Column(
        "EarningsPerShareBasic", sa.Float, nullable=False, doc="Earnings per share."
    ),
    sa.Column(
        "QuickRatio",
        sa.Float,
        nullable=False,
        doc="Current assets minus inventory over current liabilities.",
    ),
    sa.Column(
        "ReturnOnAssets",
        sa.Float,
        nullable=False,
        doc="Net income/loss over total assets.",
    ),
    sa.Column(
        "ReturnOnEquity",
        sa.Float,
        nullable=False,
        doc="Net income/loss over stockholder's equity.",
    ),
    sa.Column(
        "WorkingCapitalRatio",
        sa.Float,
        nullable=False,
        doc="Current assets over current liabilities.",
    ),
)
"""SQL table for storing refined data as managed by
:data:`finagg.sec.feat.quarterly` (an alias for
:class:`finagg.sec.feat.Quarterly`).

:meta hide-value:
"""

normalized_quarterly = sa.Table(
    "sec.refined.quarterly.normalized",
    metadata,
    sa.Column(
        "cik",
        sa.String,
        sa.ForeignKey(submissions.c.cik, ondelete="CASCADE"),
        primary_key=True,
        doc="Unique company ticker.",
    ),
    sa.Column("filed", sa.String, nullable=False, doc="Filing date."),
    sa.Column("fy", sa.Integer, primary_key=True, doc="Fiscal year the value is for."),
    sa.Column(
        "fp",
        sa.String,
        primary_key=True,
        doc="Fiscal period the value is for (e.g., Q1 or FY).",
    ),
    sa.Column(
        "NORM(LOG_CHANGE(Assets))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's total assets between quarters "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(AssetsCurrent))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's current assets between quarters "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(CommonStockSharesOutstanding))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's stock shares outstanding between"
            " quarters normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(InventoryNet))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's inventory between quarters "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(Liabilities))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's total liabilities between quarters "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(LiabilitiesCurrent))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's current liabilities between quarters "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(LOG_CHANGE(StockholdersEquity))",
        sa.Float,
        nullable=False,
        doc=(
            "Logarithmic change in a company's stockholder's equity between "
            "quarters normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(AssetCoverageRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Total assets minus short term liabilities over total liabilities "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(BookRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Total assets minus total liabilities over outstanding shares "
            "normalized against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(DebtEquityRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Total liabilities over stock holder's equity normalized against "
            "the company's industry."
        ),
    ),
    sa.Column(
        "NORM(EarningsPerShareBasic)",
        sa.Float,
        nullable=False,
        doc="Earnings per share normalized against the company's industry.",
    ),
    sa.Column(
        "NORM(QuickRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Current assets minus inventory over current liabilities normalized "
            "against the company's industry."
        ),
    ),
    sa.Column(
        "NORM(ReturnOnAssets)",
        sa.Float,
        nullable=False,
        doc=(
            "Net income/loss over total assets normalized against the company's"
            " industry."
        ),
    ),
    sa.Column(
        "NORM(ReturnOnEquity)",
        sa.Float,
        nullable=False,
        doc=(
            "Net income/loss over stockholder's equity normalized against the company's"
            " industry."
        ),
    ),
    sa.Column(
        "NORM(WorkingCapitalRatio)",
        sa.Float,
        nullable=False,
        doc=(
            "Current assets over current liabilities normalized against the company's"
            " industry."
        ),
    ),
)
"""SQL table for storing refined data as managed by
:attr:`finagg.sec.feat.Quarterly.normalized` (an alias for
:class:`finagg.sec.feat.NormalizedQuarterly`).

:meta hide-value:
"""


def get_cik(ticker: str, /, *, engine: None | Engine = None) -> str:
    """Use raw SQL data to find a company's SEC CIK from its ticker symbol.

    This is the preferred method for getting a company's SEC CIK if raw SQL
    data is installed. This method is a convenience over
    :meth:`finagg.sec.api.get_cik` for repeatedly getting company SEC
    CIKs without having to query the SEC EDGAR API. Use
    :meth:`finagg.sec.api.get_cik` if you want to get a company's ticker
    symbol without installing or accessing locally installed raw SQL data.

    Args:
        ticker: A company's ticker symbol.
        engine: Feature store database engine. Defaults to the engine
            at :data:`finagg.backend.engine`.

    Returns:
        The company's corresponding SEC CIK.

    Examples:
        Get Apple's SEC CIK from its ticker.

        >>> finagg.sec.sql.get_cik("AAPL") == "0000320193"
        True

    """
    engine = engine or backend.engine
    if not sa.inspect(engine).has_table(submissions.name):
        submissions.create(engine)
    with engine.begin() as conn:
        (cik,) = conn.execute(
            sa.select(submissions.c.cik).where(submissions.c.ticker == ticker)
        ).one()
    return str(cik)


def get_metadata(
    *, cik: None | str = None, ticker: None | str = None, engine: None | Engine = None
) -> dict[str, Any]:
    """Return a company's metadata (its SEC CIK, ticker, name, and industry
    code) from its SEC CIK or its ticker symbol.

    A convenient method for getting a company's metadata using raw SQL
    data. This method is a convenience over
    :data:`finagg.sec.api.submissions` for repeatedly getting company
    metadata without having to query the SEC EDGAR API. Use
    :data:`finagg.sec.api.submissions` if you want to get a company's
    metadata without installing or accessing locally installed raw SQL data.

    Args:
        cik: Company SEC CIK. Mutually exclusive with ``ticker``.
        ticker: Company ticker. Mutually exclusive with ``cik``.
        engine: Feature store database engine. Defaults to the engine
            at :data:`finagg.backend.engine`.

    Returns:
        A company's metadata as a dictionary.

    Examples:
        >>> finagg.sec.sql.get_metadata(ticker="MSFT")
        {'cik': '0000789019', 'ticker': 'MSFT', 'name': 'microsoft corp', 'sic': '7372'}

    """
    engine = engine or backend.engine
    if not sa.inspect(engine).has_table(submissions.name):
        submissions.create(engine)

    if bool(cik) == bool(ticker):
        raise ValueError("Must provide a `cik` or a `ticker`.")

    if ticker:
        cik = str(get_cik(ticker, engine=engine))

    with engine.begin() as conn:
        row = conn.execute(
            sa.select(
                submissions.c.cik,
                submissions.c.ticker,
                submissions.c.name,
                submissions.c.sic,
            ).where(submissions.c.cik == cik)
        ).one()
    return row._asdict()


def get_ticker(cik: str, /, *, engine: None | Engine = None) -> str:
    """Use raw SQL data to find a company's ticker from its SEC CIK.

    This is the preferred method for getting a company's ticker if raw SQL
    data is installed. This method is a convenience over
    :meth:`finagg.sec.api.get_ticker` for repeatedly getting company tickers
    without having to query the SEC EDGAR API. Use
    :meth:`finagg.sec.api.get_ticker` if you want to get a company's
    ticker symbol without installing or accessing locally installed raw
    SQL data.

    Args:
        cik: A company's SEC CIK.
        engine: Feature store database engine. Defaults to the engine
            at :data:`finagg.backend.engine`.

    Returns:
        The company's corresponding ticker symbol.

    Examples:
        Get Apple's ticker from its SEC CIK.

        >>> finagg.sec.sql.get_ticker("0000320193") == "AAPL"
        True

    """
    engine = engine or backend.engine
    if not sa.inspect(engine).has_table(submissions.name):
        submissions.create(engine)
    with engine.begin() as conn:
        (ticker,) = conn.execute(
            sa.select(submissions.c.ticker).where(submissions.c.cik == cik)
        ).one()
    return str(ticker)


def get_tickers_in_industry(
    *,
    ticker: None | str = None,
    code: None | str = None,
    level: Literal[2, 3, 4] = 2,
    engine: None | Engine = None,
) -> set[str]:
    """Get a set of tickers that all share the same industry using raw SQL data.

    This method is convenient for finding tickers within the same
    industry so they can be compared. A related and common pattern is to use
    :attr:`finagg.sec.feat.Quarterly.normalized` to get industry-normalized
    features for a particular company. Similar to other methods in this
    submodule, this will only return tickers that have raw SQL data associated
    with them.

    Args:
        ticker: Company ticker. Lookup the industry associated
            with this company. Mutually exclusive with ``code``.
        code: Industry SIC code to use for industry lookup.
            Mutually exclusive with ``ticker``.
        level: Industry level to find tickers within.
            The industry used according to ``ticker`` or ``code``
            is subsampled according to this value. Options include:

                - 2 = major group (e.g., furniture and fixtures)
                - 3 = industry group (e.g., office furnitures)
                - 4 = industry (e.g., wood office furniture)

        engine: Feature store database engine. Defaults to the engine
            at :data:`finagg.backend.engine`.

    Returns:
        A set of tickers that all share the same industry as denoted by
        another ticker (using the ``ticker`` arg) or an industry SIC code
        (using the ``code`` arg).

    Examples:
        >>> "ETSY" in finagg.sec.sql.get_tickers_in_industry(ticker="MSFT")
        True

    """
    engine = engine or backend.engine
    if not sa.inspect(engine).has_table(submissions.name):
        submissions.create(engine)
    with engine.begin() as conn:
        if ticker:
            (sic,) = conn.execute(
                sa.select(submissions.c.sic).where(submissions.c.ticker == ticker)
            ).one()
            code = str(sic)[:level]
        elif code:
            code = str(code)[:level]
        else:
            raise ValueError("Must provide a `ticker` or `code`.")

        tickers = (
            conn.execute(
                sa.select(submissions.c.ticker).where(
                    submissions.c.sic.startswith(code)
                )
            )
            .scalars()
            .all()
        )
    return set(tickers)



"""Definitions related to tracking an investment portfolio of cash and stocks.
Underlying arithmetic uses exact decimal representations for max precision.

"""

from decimal import Decimal
from functools import total_ordering


@total_ordering
class Position:
    """A position in holding a security.

    Args:
        cost: Initial purchase cost.
        quantity: Number of shares held at ``cost``.

    """

    # Average dollar cost for each share in the position.
    _average_cost_basis: Decimal

    # Current number of shares owned in the position.
    _quantity: Decimal

    # Total dollar cost for all shares in the position.
    # The amount of dollars or cash invested in this security.
    _total_cost_basis: Decimal

    def __init__(self, cost: float, quantity: float, /) -> None:
        self._average_cost_basis = Decimal(cost)
        self._quantity = Decimal(quantity)
        self._total_cost_basis = self._average_cost_basis * self._quantity

    def __eq__(self, __o: object) -> bool:
        """Compare the position's cost basis."""
        if not isinstance(__o, float | Position):
            raise TypeError(
                "Can only compare "
                f"{self.__class__.__name__} to [{float.__name__}, {Position.__name__}]"
            )

        if isinstance(__o, Position):
            return self._average_cost_basis == __o._average_cost_basis

        return self._average_cost_basis == __o

    def __lt__(self, __o: object) -> bool:
        """Compare the position's cost basis."""
        if not isinstance(__o, float | Position):
            raise TypeError(
                "Can only compare "
                f"{self.__class__.__name__} to [{float.__name__}, {Position.__name__}]"
            )

        if isinstance(__o, Position):
            return self._average_cost_basis < __o._average_cost_basis

        return self._average_cost_basis < __o

    @property
    def average_cost_basis(self) -> float:
        """Average dollar cost for each share in the position."""
        return float(self._average_cost_basis)

    def buy(self, cost: float, quantity: float, /) -> float:
        """Buy ``quantity`` of the position for ``cost``.

        Args:
            cost: Cost to buy at.
            quantity: Number of shares to buy.

        Returns:
            Value of the bought position.

        Examples:
            >>> from finagg.portfolio import Position
            >>> pos = Position(100.0, 1)
            >>> pos.buy(50.0, 1)
            50.0
            >>> pos.total_cost_basis
            150.0
            >>> pos.average_cost_basis
            75.0
            >>> pos.quantity
            2.0

        """
        exact_cost = Decimal(cost)
        exact_quantity = Decimal(quantity)
        self._quantity += exact_quantity
        self._total_cost_basis = self._total_cost_basis + exact_cost * exact_quantity
        self._average_cost_basis = self._total_cost_basis / self._quantity
        return float(exact_cost * exact_quantity)

    @property
    def quantity(self) -> float:
        """Current number of shares owned in the position."""
        return float(self._quantity)

    def sell(self, cost: float, quantity: float, /) -> float:
        """Sell ``quantity`` of the position for ``cost``.

        Args:
            cost: Cost to sell at.
            quantity: Number of shares to sell.

        Returns:
            Value of the sold position.

        Raises:
            `ValueError`: If there aren't enough shares
                to sell in the position.

        Examples:
            >>> from finagg.portfolio import Position
            >>> pos = Position(100.0, 2)
            >>> pos.sell(50.0, 1)
            50.0
            >>> pos.total_cost_basis
            100.0
            >>> pos.average_cost_basis
            100.0
            >>> pos.quantity
            1.0

        """
        exact_cost = Decimal(cost)
        exact_quantity = Decimal(quantity)
        if self._quantity < exact_quantity:
            raise ValueError("Invalid order - not enough shares.")
        self._quantity -= exact_quantity
        self._total_cost_basis = self._average_cost_basis * self._quantity
        return float(exact_cost * exact_quantity)

    @property
    def total_cost_basis(self) -> float:
        """Total dollar cost for all shares in the position. The amount of
        dollars or cash invested in this security.

        """
        return float(self._total_cost_basis)

    def total_dollar_change(self, cost: float, /) -> float:
        """Compute the total dollar change relative to the average
        cost basis and the current value of the security.

        Args:
            cost: Current value of one share.

        Returns:
            Total dollar change in value.

        Examples:
            >>> from finagg.portfolio import Position
            >>> pos = Position(100.0, 1)
            >>> pos.total_dollar_change(50.0)
            -50.0

        """
        return float((Decimal(cost) - self._average_cost_basis) * self._quantity)

    def total_log_change(self, cost: float, /) -> float:
        """Compute the total log change relative to the average
        cost basis and the current value of the security.

        Args:
            cost: Current value of one share.

        Returns:
            Total log change in value. Negative indicates loss
            in value, positive indicates gain in value.

        Examples:
            >>> from finagg.portfolio import Position
            >>> pos = Position(100.0, 1)
            >>> pos.total_log_change(50.0)
            -0.6931471805599453

        """
        return float(Decimal(cost).ln() - self._average_cost_basis.ln())

    def total_percent_change(self, cost: float, /) -> float:
        """Compute the total percent change relative to the average
        cost basis and the current value of the security.

        Args:
            cost: Current value of one share.

        Returns:
            Total percent change in value. Negative indicates loss
            in value, positive indicates gain in value.

        Examples:
            >>> from finagg.portfolio import Position
            >>> pos = Position(100.0, 1)
            >>> pos.total_percent_change(50.0)
            -0.5

        """
        return float((Decimal(cost) / self._average_cost_basis) - 1)


class Portfolio:
    """A collection of cash and security positions.

    Args:
        cash: Starting cash position.

    """

    # Total liquid cash on-hand.
    _cash: Decimal

    # Total cash deposited since starting the portfolio.
    _total_deposits: Decimal

    # Total cash withdrawn since starting the portfolio.
    _total_withdrawals: Decimal

    #: Existing positions for each security.
    positions: dict[str, Position]

    def __init__(self, cash: float, /) -> None:
        self._cash = Decimal(cash)
        self._total_deposits = self._cash
        self._total_withdrawals = Decimal(0)
        self.positions = {}

    def __contains__(self, symbol: str) -> bool:
        """Return whether the portfolio contains a position
        in ``symbol``.

        """
        return symbol in self.positions

    def __getitem__(self, symbol: str) -> Position:
        """Return the portfolio's position in the security
        identified by `symbol`.

        """
        return self.positions[symbol]

    @property
    def cash(self) -> float:
        """Total liquid cash on-hand."""
        return float(self._cash)

    def buy(self, symbol: str, cost: float, quantity: float, /) -> float:
        """Buy ``quantity`` of security with ``symbol`` for ``cost``.

        Args:
            symbol: Security ticker.
            cost: Cost to buy the symbol at.
            quantity: Number of shares to purchase.

        Returns:
            Value of the symbol's bought position in the
            portfolio.

        Raises:
            `ValueError`: If the portfolio doesn't have enough cash
                to execute the buy order.

        Examples:
            >>> from finagg.portfolio import Portfolio
            >>> port = Portfolio(1000.0)
            >>> port.buy("AAPL", 100.0, 1)
            100.0
            >>> pos = port["AAPL"]
            >>> pos.total_cost_basis
            100.0
            >>> pos.average_cost_basis
            100.0
            >>> pos.quantity
            1.0

        """
        current_value = Decimal(cost) * Decimal(quantity)
        if self._cash < current_value:
            raise ValueError("Invalid order - not enough cash.")
        self._cash -= current_value
        if symbol not in self.positions:
            self.positions[symbol] = Position(cost, quantity)
            return float(current_value)
        else:
            return self.positions[symbol].buy(cost, quantity)

    def deposit(self, cash: float, /) -> float:
        """Deposit more cash into the portfolio.

        Args:
            cash: Cash to deposit.

        Returns:
            Total cash in the portfolio.

        """
        exact_cash = Decimal(cash)
        self._cash += exact_cash
        self._total_deposits += exact_cash
        return float(self._cash)

    def sell(self, symbol: str, cost: float, quantity: float, /) -> float:
        """Sell ``quantity`` of security with `symbol` for ``cost``.

        Args:
            symbol: Security ticker.
            cost: Cost to sell the symbol at.
            quantity: Number of shares to sell.

        Returns:
            Value of the symbol's sold position in the
            portfolio.

        Examples:
            >>> from finagg.portfolio import Portfolio
            >>> port = Portfolio(1000.0)
            >>> port.buy("AAPL", 100.0, 2)
            200.0
            >>> port.sell("AAPL", 50.0, 1)
            50.0
            >>> pos = port["AAPL"]
            >>> pos.total_cost_basis
            100.0
            >>> pos.average_cost_basis
            100.0
            >>> pos.quantity
            1.0

        """
        current_value = self.positions[symbol].sell(cost, quantity)
        if not self.positions[symbol]._quantity:
            self.positions.pop(symbol)
        self._cash += Decimal(cost) * Decimal(quantity)
        return float(current_value)

    @property
    def total_deposits(self) -> float:
        """Total cash deposited since starting the portfolio."""
        return float(self._total_deposits)

    def total_dollar_change(self, costs: dict[str, float], /) -> float:
        """Compute the total dollar change relative to the total
        deposits made into the portfolio.

        Args:
            costs: Mapping of symbol to its current value of one share.

        Returns:
            Total dollar change in value.

        Examples:
            >>> from finagg.portfolio import Portfolio
            >>> port = Portfolio(1000.0)
            >>> port.buy("AAPL", 100.0, 1)
            100.0
            >>> port.total_dollar_change({"AAPL": 50.0})
            -50.0

        """
        return float(Decimal(self.total_dollar_value(costs)) - self._total_deposits)

    def total_dollar_value(self, costs: dict[str, float], /) -> float:
        """Compute the total dollar value of the portfolio.

        Args:
            costs: Mapping of symbol to its current value of one share.

        Returns:
            Total dollar value.

        Examples:
            >>> from finagg.portfolio import Portfolio
            >>> port = Portfolio(1000.0)
            >>> port.buy("AAPL", 100.0, 1)
            100.0
            >>> port.total_dollar_value({"AAPL": 50.0})
            950.0

        """
        total_dollar_value = self._cash
        for symbol, cost in costs.items():
            if symbol in self.positions:
                total_dollar_value += Decimal(cost) * self.positions[symbol]._quantity
        return float(total_dollar_value)

    def total_log_change(self, costs: dict[str, float], /) -> float:
        """Compute the total log change relative to the total
        deposits made into the portfolio.

        Args:
            costs: Mapping of symbol to its current value of one share.

        Returns:
            Total log change in value. Negative indicates loss
            in value, positive indicates gain in value.

        Examples:
            >>> from finagg.portfolio import Portfolio
            >>> port = Portfolio(1000.0)
            >>> port.buy("AAPL", 100.0, 1)
            100.0
            >>> port.total_log_change({"AAPL": 50.0})
            -0.051293294387550536

        """
        return float(
            Decimal(self.total_dollar_value(costs)).ln() - self._total_deposits.ln()
        )

    def total_percent_change(self, costs: dict[str, float], /) -> float:
        """Compute the total percent change relative to the total
        deposits made into the portfolio.

        Args:
            costs: Mapping of symbol to its current value of one share.

        Returns:
            Total percent change in value. Negative indicates loss
            in value, positive indicates gain in value.

        Examples:
            >>> from finagg.portfolio import Portfolio
            >>> port = Portfolio(1000.0)
            >>> port.buy("AAPL", 100.0, 1)
            100.0
            >>> port.total_percent_change({"AAPL": 50.0})
            -0.05

        """
        return float(
            (Decimal(self.total_dollar_value(costs)) / self._total_deposits) - 1
        )

    @property
    def total_withdrawals(self) -> float:
        """Total cash withdrawn since starting the portfolio."""
        return float(self._total_withdrawals)

    def withdraw(self, cash: float, /) -> float:
        """Withdraw cash from the portfolio.

        Args:
            cash: Cash to withdraw.

        Returns:
            Total cash in the portfolio.

        Raises:
            `ValueError`: If the portfolio doesn't have at least ``cash``
                liquid cash to withdraw.

        Examples:
            >>> from finagg.portfolio import Portfolio
            >>> port = Portfolio(1000.0)
            >>> port.withdraw(100.0)
            900.0

        """
        exact_cash = Decimal(cash)
        if self._cash < exact_cash:
            raise ValueError("Not enough cash to withdraw.")
        self._cash -= exact_cash
        self._total_withdrawals += exact_cash
        return float(self._cash)



"""Refined Yahoo! Finance features (features aggregated from raw tables)."""

import logging
import multiprocessing as mp

import numpy as np
import pandas as pd
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound, NoSuchTableError

from ... import backend, utils
from .. import api, sql
from . import _raw

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class Daily:
    """Methods for gathering daily stock data features from Yahoo! finance.

    The module variable :data:`finagg.yfinance.feat.daily` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    Examples:
        It doesn't matter which data source you use to gather features.
        They all return equivalent dataframes.

        >>> df1 = finagg.yfinance.feat.daily.from_api("AAPL").head(5)
        >>> df2 = finagg.yfinance.feat.daily.from_raw("AAPL").head(5)
        >>> df3 = finagg.yfinance.feat.daily.from_refined("AAPL").head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
        >>> pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)

    """

    @classmethod
    def _from_raw_for_update(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Helper for getting raw features for the update method."""
        engine = engine or backend.engine
        if start is None:
            with engine.begin() as conn:
                (start,) = conn.execute(
                    sa.select(sa.func.max(sql.daily.c.date)).where(
                        sql.daily.c.ticker == ticker
                    )
                ).one()
        return cls.from_raw(ticker, start=start, end=end, engine=engine)

    @classmethod
    def _normalize(cls, df: pd.DataFrame, /) -> pd.DataFrame:
        """Normalize daily features columns."""
        df = df.drop(columns=["ticker"]).set_index("date").sort_index()
        df = df.replace([-np.inf, np.inf], np.nan).ffill()
        df = utils.resolve_func_cols(sql.daily, df, drop=True, inplace=True)
        return df.dropna()

    @classmethod
    def from_api(
        cls, ticker: str, /, *, start: None | str = None, end: None | str = None
    ) -> pd.DataFrame:
        """Get daily features directly from :meth:`finagg.yfinance.api.get`.

        Args:
            ticker: Company ticker.
            start: The start date of the stock history. Defaults to the
                first recorded date.
            end: The end date of the stock history. Defaults to the
                last recorded date.

        Returns:
            Daily stock price dataframe sorted by date.

        Examples:
            >>> finagg.yfinance.feat.daily.from_api("AAPL").head(5)  # doctest: +SKIP
                        LOG_CHANGE(open)  LOG_CHANGE(high)  LOG_CHANGE(low)  LOG_CHANGE(close) ...
            date                                                                               ...
            1980-12-15         -0.049005         -0.053343        -0.053581          -0.053581 ...
            1980-12-16         -0.075870         -0.075870        -0.076231          -0.076231 ...
            1980-12-17          0.019512          0.024331         0.024450           0.024450 ...
            1980-12-18          0.028580          0.028445         0.028580           0.028580 ...
            1980-12-19          0.059239          0.058970         0.059239           0.059239 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        df = api.get(ticker, start=start, end=end)
        return cls._normalize(df)

    @classmethod
    def from_raw(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get daily features from local SQL tables.

        Args:
            ticker: Company ticker.
            start: The start date of the stock history. Defaults to the
                first recorded date.
            end: The end date of the stock history. Defaults to the
                last recorded date.
            engine: Raw store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Daily stock price dataframe sorted by date.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the
                raw SQL table.

        Examples:
            >>> finagg.yfinance.feat.daily.from_raw("AAPL").head(5)  # doctest: +SKIP
                        LOG_CHANGE(open)  LOG_CHANGE(high)  LOG_CHANGE(low)  LOG_CHANGE(close) ...
            date                                                                               ...
            1980-12-15         -0.049005         -0.053343        -0.053581          -0.053581 ...
            1980-12-16         -0.075870         -0.075870        -0.076231          -0.076231 ...
            1980-12-17          0.019512          0.024331         0.024450           0.024450 ...
            1980-12-18          0.028580          0.028445         0.028580           0.028580 ...
            1980-12-19          0.059239          0.058970         0.059239           0.059239 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.prices.name):
            sql.prices.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.prices.select().where(
                        sql.prices.c.ticker == ticker,
                        sql.prices.c.date >= start,
                        sql.prices.c.date <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No daily rows found for {ticker}.")
        return cls._normalize(df)

    @classmethod
    def from_refined(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from the feature-dedicated local SQL tables.

        This is the preferred method for accessing features for
        offline analysis (assuming data in the local SQL tables
        is installed and is up-to-date).

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Daily stock price dataframe sorted by date.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the
                refined SQL table.

        Examples:
            >>> finagg.yfinance.feat.daily.from_refined("AAPL").head(5)  # doctest: +SKIP
                        LOG_CHANGE(open)  LOG_CHANGE(high)  LOG_CHANGE(low)  LOG_CHANGE(close) ...
            date                                                                               ...
            1980-12-15         -0.049005         -0.053343        -0.053581          -0.053581 ...
            1980-12-16         -0.075870         -0.075870        -0.076231          -0.076231 ...
            1980-12-17          0.019512          0.024331         0.024450           0.024450 ...
            1980-12-18          0.028580          0.028445         0.028580           0.028580 ...
            1980-12-19          0.059239          0.058970         0.059239           0.059239 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.daily.name):
            sql.daily.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.daily.select().where(
                        sql.daily.c.ticker == ticker,
                        sql.daily.c.date >= start,
                        sql.daily.c.date <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No daily rows found for {ticker}.")
        return df.drop(columns=["ticker"]).set_index("date").sort_index()

    @classmethod
    def get_candidate_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the raw SQL table that MAY BE ELIGIBLE
        to be in the feature's refined SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that may be valid for creating daily features
            that also have at least ``lb`` rows used for constructing the
            features.

        Examples:
            >>> "AAPL" in finagg.yfinance.feat.daily.get_candidate_ticker_set()  # doctest: +SKIP
            True

        """
        return _raw.Prices.get_ticker_set(lb=lb, start=start, end=end, engine=engine)

    @classmethod
    def get_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the feature's SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that contain all the columns for creating
            daily features that also have at least ``lb`` rows.

        Examples:
            >>> "AAPL" in finagg.yfinance.feat.daily.get_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.daily.name):
            sql.daily.create(engine)
        with engine.begin() as conn:
            tickers = (
                conn.execute(
                    sa.select(sql.daily.c.ticker)
                    .where(
                        sql.daily.c.date >= start,
                        sql.daily.c.date <= end,
                    )
                    .group_by(sql.daily.c.ticker)
                    .having(sa.func.count(sql.daily.c.date) >= lb)
                )
                .scalars()
                .all()
            )
        return set(tickers)

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from the
        raw SQL tables, transforming them into daily features, and then writing
        to the refined daily SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.indices.api.get_ticker_set`.
            processes: Number of background processes to run in parallel
                when processing ``tickers``.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's refined SQL table.

        """
        tickers = tickers or cls.get_candidate_ticker_set()
        if not tickers:
            logger.info(
                "Skipping finagg.yfinance.feat.daily installation because no tickers"
                " were provided or no tickers were found with prerequisite data (i.e.,"
                " finagg.yfinance.feat.prices data)"
            )
            return 0

        engine = engine or backend.engine
        if recreate_tables or not sa.inspect(engine).has_table(sql.daily.name):
            sql.daily.drop(engine, checkfirst=True)
            sql.daily.create(engine)

        return utils._install(
            cls.from_raw,
            cls.to_refined,
            logger,
            list(tickers),
            engine,
            desc="Installing refined Yahoo! Finance daily data",
            processes=processes,
        )

    @classmethod
    def to_refined(
        cls,
        ticker: str,
        df: pd.DataFrame,
        /,
        *,
        engine: None | Engine = None,
    ) -> int:
        """Write the given dataframe to the refined feature table
        while using the ticker ``ticker``.

        Args:
            ticker: Company ticker.
            df: Dataframe to store as rows in a local SQL table.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.daily.name):
            sql.daily.create(engine)
        df = df.reset_index("date")
        df["ticker"] = ticker
        with engine.begin() as conn:
            conn.execute(sql.daily.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df.index)

    @classmethod
    def update(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
    ) -> int:
        """Update data associated with ``tickers`` by pulling data from the
        raw SQL tables, transforming them into daily features, and then writing
        to the refined daily SQL table.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.yfinance.feat.Daily.get_ticker_set`.
            processes: Number of background processes to run in parallel
                when processing ``tickers``.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the feature's refined SQL table.

        Raises:
            `NoSuchTableError`: If the table associated with this feature
                set update does not exist.

        """
        tickers = tickers or cls.get_ticker_set()
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.prices.name):
            raise NoSuchTableError(f"{sql.daily.name} table does not exist.")

        return utils._install(
            cls._from_raw_for_update,
            cls.to_refined,
            logger,
            list(tickers),
            engine,
            desc="Updating refined Yahoo! Finance daily data",
            processes=processes,
        )



"""Abstract FRED API definition."""

import os
from abc import ABC, abstractmethod
from datetime import timedelta
from functools import partial
from typing import Any, ClassVar

import pandas as pd
import requests
import requests_cache

from ... import backend, ratelimit

if backend.disable_http_cache:
    session = requests.Session()
else:
    session = requests_cache.CachedSession(
        str(backend.http_cache_path),
        ignored_parameters=["api_key", "file_type"],
        expire_after=timedelta(weeks=1),
    )


class API(ABC):
    """Abstract FRED API."""

    #: Request API URL.
    url: ClassVar[str]

    @classmethod
    @abstractmethod
    def get(cls, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """Main dataset API method."""


@ratelimit.guard([ratelimit.RequestLimit(120, timedelta(minutes=1))])
def get(url: str, /, **kwargs: Any) -> requests.Response:
    """Main API get function used by all `Dataset.get` methods.

    Also formats FRED API parameters for convenience.

    Args:
        url: Request URL.
        **kwargs: Mapping of request parameter name to their value.

    Returns:
        A valid FRED API response.

    """
    params = pformat(**kwargs)
    if params.pop("cache", True):
        response = session.get(url, params=params)
    else:
        response = requests.get(url, params=params)
    response.raise_for_status()
    return response


def maybe_paginate(data_key: str, url: str, /, **kwargs: Any) -> pd.DataFrame:
    """Do pagination for API get functions that support pagination (if
    pagination is enabled).

    Args:
        data_key: Key from the response that the underlying data resides in.
        url: Request URL.
        **kwargs: Mapping of request parameter name to their value.

    Returns:
        A dataframe containing (possibly all) data results.

    """
    do_paginate = kwargs.pop("paginate", False)
    offset = kwargs.pop("offset", 0)
    getter = partial(get, url, **kwargs)

    buffer: list[dict[str, Any]] = []
    batch_size = kwargs.get("limit", 1000)
    data = {"count": offset}
    i = offset
    while not buffer or i <= data["count"]:
        data = getter(offset=i).json()
        buffer.extend(data[data_key])
        if not do_paginate:
            break
        i += batch_size
    return pd.DataFrame(buffer)


def pformat(**kwargs: Any) -> dict[str, Any]:
    """FRED API parameter formatting.

    Args:
        **kwargs: All possible FRED API parameters.

    Returns:
        Mapping of request parameter name to their value.

    """
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    for k in ("observation_start", "observation_end", "realtime_start", "realtime_end"):
        if k in kwargs:
            match kwargs[k]:
                case 0:
                    kwargs[k] = "1776-07-04"
                case -1:
                    kwargs[k] = "9999-12-31"

    for k in ("exclude_tag_names", "tag_names"):
        if k in kwargs:
            v = kwargs[k]
            if isinstance(v, str):
                v = [v]
            kwargs[k] = ";".join(v)

    for k in ("include_observation_values", "include_release_dates_with_no_data"):
        if k in kwargs:
            kwargs[k] = "true" if kwargs[k] else "false"

    for k in ("search_text", "series_search_text", "tag_search_text"):
        if k in kwargs:
            v = kwargs[k]
            if isinstance(v, str):
                v = [v]
            kwargs[k] = "+".join(v)

    if "vintage_dates" in kwargs:
        k = "vintage_dates"
        v = kwargs[k]
        if isinstance(v, str):
            v = [v]
        kwargs[k] = ",".join(v)

    api_key = kwargs.pop("api_key", None) or os.environ.get("FRED_API_KEY", None)
    if not api_key:
        raise RuntimeError(
            "No FRED API key found. "
            "Pass the API key to the API directly, or "
            "set the `FRED_API_KEY` environment variable."
        )
    kwargs.update({"api_key": api_key, "file_type": "json"})
    return kwargs



"""SEC CLI and tools."""

import logging
import os

import click

from .. import utils

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


@click.group(help="Bureau of Economic Analysis (BEA) tools.")
def entry_point() -> None:
    ...


@entry_point.command(help="Set the BEA API key.")
def install() -> int:
    if "BEA_API_KEY" not in os.environ:
        api_key = input(
            "Enter your BEA API key below.\n\n"
            "You can request a BEA API key at\n"
            "https://apps.bea.gov/api/signup/.\n\n"
            "BEA API key (leave blank and hit ENTER to skip): "
        ).strip()
        if not api_key:
            logger.warning(
                "An empty BEA API key was given. Skipping finagg.bea installation."
            )
            return 0
        p = utils.setenv("BEA_API_KEY", api_key)
        logger.info(f"BEA API key written to {p}")
    else:
        logger.info("BEA API key found in the environment")
    return 0
